#!/usr/bin/env python
#
# filetreepanel.py - The FileTreePanel
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`FileTreePanel` class, which can be used
to browse the contents of structured directories which are described with
a :mod:`.filetree`.

See also the :mod:`.filetreemanager` module, which contains the logic for
generating the file list.
"""


import os.path   as op
import itertools as it
import              time
import              logging
import              collections

import              wx

import fsl.utils.idle                   as idle
import fsl.utils.filetree               as filetree
import fsl.utils.settings               as fslsettings
import fsleyes_widgets.utils.overlay    as fwoverlay
import fsleyes_widgets.widgetlist       as wlist
import fsleyes_widgets.widgetgrid       as wgrid
import fsleyes_widgets.elistbox         as elb

import fsleyes.strings                  as strings
import fsleyes.controls.controlpanel    as ctrlpanel
import fsleyes.controls.filetreemanager as ftmanager


log = logging.getLogger(__name__)

NONELBL    = strings.labels['VariablePanel.value.none']
ANYLBL     = strings.labels['VariablePanel.value.any']
ALLLBL     = strings.labels['VariablePanel.value.all']
PRESENTLBL = strings.labels['FileListPanel.present']


BUILTIN_TREE_FILTER = ['BedpostX', 'Diffusion', 'HCP_Surface', 'ProbtrackX',
                       'bet',      'dti',       'eddy',        'epi_reg',
                       'fast',     'topup',     'feat_reg',    'feat_stats']
"""Built-in ``.tree`` files with a name in this list are hidden from the
:class:`FileTreePanel` interface. These trees are not very useful for our
purposes of navigating multi-subject data directories.
"""


class FileTreePanel(ctrlpanel.ControlPanel):
    """The ``FileTreePanel`` can be used to browse the contents of structured
    directories which are described with a :mod:`.filetree`.

    The user needs to select a data directory, and a file tree. The file tree
    can be selected either from the drop down list of built-in trees, or a
    custom tree file can be selected.

    Once the user has selected a file tree and a data directory, the
    :class:`FileTypePanel` and :class:`VariablePanel` will be populated,
    allowing the user to choose which file types to display, and how to
    arrange them.

    When the user has selected some file types, the :class:`FileListPanel`
    will display a grid containing all of the matching files that exist in the
    directory. The user can select a row to view the relevant files.

    The :class:`.FileTreeManager` handles the logic of working with the
    :class:`.FileTree` and of displaying overlays.
    """


    customTrees = []
    """Whenever the user loads a custom tree file, its path is added to this
    list, so that the tree file dropdown box can be populated with previously
    loaded tree files.
    """


    def __init__(self, parent, overlayList, displayCtx, frame):
        """Create a ``FileTreePanel``.

        :arg parent:      The :mod:`wx` parent object.

        :arg overlayList: The :class:`.OverlayList` instance.

        :arg displayCtx:  The :class:`.DisplayContext` instance.

        :arg frame:       The :class:`.FSLeyesFrame` instance.
        """
        ctrlpanel.ControlPanel.__init__(
            self, parent, overlayList, displayCtx, frame)

        self.__tree  = None
        self.__query = None
        self.__mgr   = None

        self.__loadDir      = wx.Button(self)
        self.__customTree   = wx.Button(self)
        self.__treeChoice   = wx.Choice(self)
        self.__save         = wx.Button(self)
        self.__mainSplitter = wx.SplitterWindow(
            self, style=wx.SP_LIVE_UPDATE | wx.SP_BORDER)
        self.__leftPanel    = wx.Panel(self.__mainSplitter)
        self.__rightPanel   = wx.Panel(self.__mainSplitter)
        self.__notesLabel   = wx.StaticText(self.__leftPanel)
        self.__notesChoice  = wx.Choice(self.__leftPanel)
        self.__dirName      = wx.StaticText(
            self.__rightPanel, style=wx.ST_ELLIPSIZE_MIDDLE)

        self.__varPanel     = VariablePanel(self.__leftPanel,  self)
        self.__fileTypes    = FileTypePanel(self.__leftPanel,  self)
        self.__fileList     = FileListPanel(self.__rightPanel, self)

        self.__mainSplitter.SetMinimumPaneSize(50)
        self.__mainSplitter.SplitVertically(
            self.__leftPanel, self.__rightPanel)
        self.__mainSplitter.SetSashGravity(0.3)

        # Build a list of all built-in filetrees,
        # along with any custom ones that have been
        # previously loaded. We hide the BUILTIN
        # ones
        def filter(tf):
            name = op.splitext(op.basename(tf))[0]
            return name not in BUILTIN_TREE_FILTER

        treefiles   = [tf for tf in filetree.list_all_trees() if filter(tf)]
        treefiles   = list(sorted(treefiles))
        treefiles  += FileTreePanel.customTrees
        treefiles   = [op.abspath( tf) for tf in treefiles]
        treelabels  = [op.basename(tf) for tf in treefiles]

        for f, l in zip(treefiles, treelabels):
            self.__treeChoice.Append(l, clientData=f)

        notesChoices = list(strings.choices[self, 'notes'].items())
        for key, label in notesChoices:
            self.__notesChoice.Append(label, clientData=key)

        self.__treeChoice .SetSelection(0)
        self.__notesChoice.SetSelection(0)
        self.__fileList   .NotesColumn(notesChoices[0][0])

        self.__loadDir   .SetLabel(strings.labels[self, 'loadDir'])
        self.__customTree.SetLabel(strings.labels[self, 'customTree'])
        self.__save      .SetLabel(strings.labels[self, 'save'])
        self.__notesLabel.SetLabel(strings.labels[self, 'notes'])

        self.__mainSizer  = wx.BoxSizer(wx.VERTICAL)
        self.__topSizer   = wx.BoxSizer(wx.HORIZONTAL)
        self.__notesSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.__leftSizer  = wx.BoxSizer(wx.VERTICAL)
        self.__rightSizer = wx.BoxSizer(wx.VERTICAL)

        self.__notesSizer.Add(self.__notesLabel,  flag=wx.EXPAND)
        self.__notesSizer.Add(self.__notesChoice, flag=wx.EXPAND, proportion=1)

        self.__leftSizer.Add(self.__fileTypes,  flag=wx.EXPAND, proportion=1)
        self.__leftSizer.Add(self.__varPanel,   flag=wx.EXPAND)
        self.__leftSizer.Add(self.__notesSizer, flag=wx.EXPAND)

        self.__rightSizer.Add(self.__dirName,  flag=wx.EXPAND)
        self.__rightSizer.Add(self.__fileList, flag=wx.EXPAND, proportion=1)

        self.__topSizer.Add(self.__treeChoice, flag=wx.EXPAND, proportion=1)
        self.__topSizer.Add(self.__loadDir,    flag=wx.EXPAND, proportion=1)
        self.__topSizer.Add(self.__customTree, flag=wx.EXPAND, proportion=1)
        self.__topSizer.Add(self.__save,       flag=wx.EXPAND, proportion=1)

        self.__mainSizer.Add(self.__topSizer,     flag=wx.EXPAND)
        self.__mainSizer.Add(self.__mainSplitter, flag=wx.EXPAND, proportion=1)

        self             .SetSizer(self.__mainSizer)
        self.__leftPanel .SetSizer(self.__leftSizer)
        self.__rightPanel.SetSizer(self.__rightSizer)

        self.__loadDir    .Bind(wx.EVT_BUTTON, self._onLoadDir)
        self.__customTree .Bind(wx.EVT_BUTTON, self._onCustomTree)
        self.__treeChoice .Bind(wx.EVT_CHOICE, self._onTreeChoice)
        self.__save       .Bind(wx.EVT_BUTTON, self._onSave)
        self.__notesChoice.Bind(wx.EVT_CHOICE, self._onNotesChoice)


    @property
    def varPanel(self):
        """Return a reference to the :class:`VariablePanel`."""
        return self.__varPanel


    @property
    def fileTypePanel(self):
        """Return a reference to the :class:`FileTypePanel`."""
        return self.__fileTypes


    @property
    def fileListPanel(self):
        """Return a reference to the :class:`FileListPanel`."""
        return self.__fileList


    @property
    def treeChoice(self):
        """Return a reference to the file tree ``wx.Choice`` widget."""
        return self.__treeChoice


    def UpdateFileList(self):
        """Called by the sub-panels when the user changes any settings.
        Re-generates the file grid.
        """

        if self.__tree is None or self.__query is None:
            return

        mgr      = self.__mgr
        flist    = self.__fileList
        ftypes   = self.__fileTypes.GetFileTypes()
        varyings = self.__varPanel.GetVaryings()
        fixed    = self.__varPanel.GetFixed()

        mgr.update(ftypes, varyings, fixed)
        flist.ResetGrid(mgr)


    def _loadTree(self, treename, dirname):
        """Called when a new tree or data directory is selected. Clears
        any previous file tree, and loads the new one. If either the tree
        or directory are ``None``, any existing file tree is cleared.

        :arg treename: File tree name or file
        :arg dirname:  Data directory
        """

        if treename is None or dirname is None:
            dirname   = None
            tree      = None
            query     = None
            mgr       = None
            allvars   = None
            filetypes = None
        else:
            tree      = filetree.FileTree.read(treename, directory=dirname)
            query     = filetree.FileTreeQuery(tree)
            mgr       = ftmanager.FileTreeManager(self.overlayList,
                                                  self.displayCtx,
                                                  query)
            allvars   = query.variables()
            allvars   = [(var, vals) for var, vals in allvars.items()]
            allvars   = collections.OrderedDict(list(sorted(allvars)))
            filetypes = list(sorted(query.templates))

        self.__tree  = tree
        self.__query = query
        self.__mgr   = mgr

        self.__dirName  .SetLabel(dirname or '')
        self.__varPanel .SetVariables(allvars)
        self.__fileTypes.SetFileTypes(filetypes)
        self.__leftPanel.Layout()

        self.UpdateFileList()


    def _getTreeChoice(self):
        """Returns the current selection of the built-in filetree drop down
        box.
        """
        idx = self.__treeChoice.GetSelection()

        if idx == wx.NOT_FOUND:
            return None

        return self.__treeChoice.GetClientData(idx)


    def _onLoadDir(self, ev=None):
        """Called when the user pushes the *load data directory* button.

        Prompts the user to select a directory, then calls the
        :meth:`__loadTree` method.
        """

        msg     = strings.messages[self, 'loadDir']
        fromDir = fslsettings.read('loadSaveOverlayDir')

        dlg = wx.DirDialog(wx.GetApp().GetTopWindow(),
                           message=msg,
                           defaultPath=fromDir,
                           style=wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST)

        if dlg.ShowModal() != wx.ID_OK:
            return

        dirname  = dlg.GetPath()
        treename = self._getTreeChoice()

        fslsettings.write('loadSaveOverlayDir', dirname)

        self._loadTree(treename, dirname)


    def _onTreeChoice(self, ev=None):
        """Called when the user changes the built-in file tree selection.
        Calls the :meth:`__loadTree` method.
        """
        dirname  = self.__dirName.GetLabel() or None
        treename = self._getTreeChoice()
        self._loadTree(treename, dirname)


    def _onCustomTree(self, ev=None):
        """Called when the user pushes the *load custom tree* button.
        Prompts the user to choose a file, then calls :meth:`__loadTree`.
        """
        msg = strings.messages[self, 'loadCustomTree']
        dlg = wx.FileDialog(wx.GetApp().GetTopWindow(),
                            message=msg,
                            style=wx.FD_OPEN)

        if dlg.ShowModal() != wx.ID_OK:
            return

        # TODO make sure tree label is unique
        choice    = self.__treeChoice
        ntrees    = choice.GetCount()
        treefile  = op.abspath(dlg.GetPath())
        treelabel = op.basename(treefile)
        existing  = [choice.GetClientData(i) for i in range(ntrees)]

        try:               index = existing.index(treefile)
        except ValueError: index = -1

        if index == -1:
            index = ntrees
            choice.Append(treelabel, clientData=treefile)

        choice.SetSelection(index)
        self._onTreeChoice()


    def _onSave(self, ev=None):
        """Called when the *save* button is pushed. Prompts the user
        for a destination, and then saves the contents of the grid.
        """

        msg = strings.messages[self, 'save']
        dlg = wx.FileDialog(
            self,
            defaultFile='notes.txt',
            message=msg,
            style=wx.FD_SAVE)

        if dlg.ShowModal() != wx.ID_OK:
            return

        path = dlg.GetPath()
        grid = self.__fileList.GridContents()

        with open(path, 'wt') as f:
            for row in grid:
                f.write('\t'.join(row) + '\n')


    def _onNotesChoice(self, ev):
        """Called when the user changes the notes column position choice.
        Calls :meth:`FileListPanel.NotesColumn` accordingly.
        """
        sel = self.__notesChoice.GetSelection()
        sel = self.__notesChoice.GetClientData(sel)
        self.__fileList.NotesColumn(sel)


class FileTypePanel(elb.EditableListBox):
    """The ``FileTypePanel`` displays a list of available file types.
    It allows the user to choose which file types should be displayed.

    Whenever the user changes the selected file types, the
    :meth:`FileTreePanel.Update` method is called.
    """

    def __init__(self, parent, ftpanel):
        """Create a ``FileTypePanel``.

        :arg parent:  ``wx`` parent object
        :arg ftpanel: The :class:`.FileTreePanel`
        """
        elb.EditableListBox.__init__(
            self, parent, vgap=5, style=(elb.ELB_NO_ADD    |
                                         elb.ELB_NO_REMOVE |
                                         elb.ELB_NO_MOVE))

        self.__ftpanel = ftpanel


    def SetFileTypes(self, filetypes):
        """Display the given list of file types.

        A check box is displayed alongside each file type allowing the
        user to toggle it on and off.

        :arg filetypes: List of file types Pass in ``None`` to clear the
                        list.
        """
        self.Clear()

        if filetypes is None:
            return

        for ftype in filetypes:
            toggle = wx.CheckBox(self)
            toggle.SetMinSize(toggle.GetBestSize())
            self.Append(ftype, extraWidget=toggle)
            toggle.Bind(wx.EVT_CHECKBOX, self.__onToggle)


    def GetFileTypes(self):
        """Returns a list of the file types that are currently selected. """

        filetypes = self.GetLabels()
        toggles   = self.GetWidgets()
        active    = []

        for ftype, tog in zip(filetypes, toggles):
            if tog.GetValue():
                active.append(ftype)

        return active


    def __onToggle(self, ev):
        """Called when a file type is toggled. Calls the
        :meth:`FileTreePanel.Update` method.
        """
        self.__ftpanel.UpdateFileList()


class VariablePanel(wx.Panel):
    """The ``VariablePanel`` displays a list of available variables, allowing
    the user to choose between:

      - Displaying each variable value on a different row (``<any>``, the
        default). These variables are referred to as *varying*.

      - Displaying all variable value on the same row (``<all>``). These
        variables are referred to as *fixed*.

      - Displaying one specific variable value. These are also included as
        *varying* variables.
    """


    def __init__(self, parent, ftpanel):
        """Create a ``VariablePanel``

        :arg parent:  ``wx`` parent object
        :arg ftpanel: The :class:`.FileTreePanel`
        """

        wx.Panel.__init__(self, parent)

        self.__ftpanel  = ftpanel
        self.__varList  = wlist.WidgetList(self)
        self.__vars     = None

        self.__sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(self.__sizer)

        self.__sizer.Add(self.__varList, flag=wx.EXPAND, proportion=1)


    def SetVariables(self, vars):
        """Set the variables to be displayed.

        :arg vars: Dict of ``{ var : [value] }`` mappings, containing
                   all variables, and all of their possible values.
        """

        self.__vars = None
        self.__varList.Clear()

        if vars is None or len(vars) == 0:
            return

        self.__vars = {}

        for var, vals in vars.items():

            if None in vals:
                vals.remove(None)
                vals = [ANYLBL, ALLLBL, NONELBL] + sorted(vals)
            else:
                vals = [ANYLBL, ALLLBL]          + sorted(vals)

            choice = wx.Choice(self.__varList, choices=vals, name=var)

            choice.SetSelection(0)
            choice.Bind(wx.EVT_CHOICE, self.__onVariable)

            self.__varList.AddWidget(choice, var)
            self.__vars[var] = vals


    def GetVaryings(self):
        """Return a dict of ``{ var : val }`` mappings containing all *varying*
        variables. The value for each variable may be one of:

         - ``'*'``, indicating that all possible values for this variable
           should be considered

         - ``None``, indicating that only instances  where this variable is
           absent should be considered.

         - A specific value, indicating that only this value should be
           considered.
        """

        queryvars = {}

        for choice in self.__varList.GetWidgets():
            var    = choice.GetName()
            validx = choice.GetSelection()
            val    = self.__vars[var][validx]

            if   val == ALLLBL:  continue
            elif val == ANYLBL:  queryvars[var] = '*'
            elif val == NONELBL: queryvars[var] = None
            else:                queryvars[var] = val

        return queryvars


    def GetFixed(self):
        """Returns a list containing the names of all *fixed* variables. """

        fixedvars = []

        for choice in self.__varList.GetWidgets():
            var    = choice.GetName()
            validx = choice.GetSelection()
            val    = self.__vars[var][validx]

            if val == ALLLBL:
                fixedvars.append(var)

        return fixedvars


    def __onVariable(self, ev):
        """Called when the user changes a variable setting. Calls the
        :meth:`FileTreePanel.Update` method.
        """
        self.__ftpanel.UpdateFileList()


class FileListPanel(wx.Panel):
    """The ``FileListPanel`` displays a grid of filetree variable values
    and file types, allowing the user to step through the files in the data
    directory.


    The user can drag varying variable columns to re-order them - this will
    trigger a call to :meth:`.FileTreeManager.reorder`.


    A *Notes* column contains text controls in each row, allowing the user to
    add notes. The position of this column can be set to either left or right
    of the fixed variable columns via the :meth:`NotesColumn` method.
    """


    def __init__(self, parent, ftpanel, notes='right'):
        """Create a ``FileListPanel``.

        :arg parent:  ``wx`` parent object
        :arg ftpanel: The :class:`.FileTreePanel`
        :arg notes:   Location of the *Notes* column - one of:

                        - ``'right'`` - right-most column (default)
                        - ``'left'``  - left-most column, after varying
                          columns

        The *Notes* column location can be changed later via the
        :meth:`NotesColumn` method.
        """

        if notes not in ('left', 'right'):
            raise ValueError('Invalid value for notes: {}'.format(notes))

        wx.Panel.__init__(self, parent)

        self.__ftpanel = ftpanel
        self.__notes   = notes
        self.__mgr     = None
        self.__sizer   = wx.BoxSizer(wx.VERTICAL)
        self.__grid    = wgrid.WidgetGrid(
            self,
            style=(wx   .HSCROLL            |
                   wx   .VSCROLL            |
                   wgrid.WG_KEY_NAVIGATION  |
                   wgrid.WG_SELECTABLE_ROWS |
                   wgrid.WG_DRAGGABLE_COLUMNS))

        self.__sizer.Add(self.__grid, flag=wx.EXPAND, proportion=1)
        self.SetSizer(self.__sizer)

        self.__grid.Bind(wgrid.EVT_WG_SELECT,  self.__onSelect)
        self.__grid.Bind(wgrid.EVT_WG_REORDER, self.__onReorder)


    def ResetGrid(self, mgr):
        """Clear and re-populate the file tree grid.

        :arg mgr: The :class:`.FileTreeManager` from which the variable
                  and file information is retrieved.
        """

        self.__mgr = mgr
        varcols    = mgr.varcols
        fixedcols  = mgr.fixedcols
        fgroups    = mgr.filegroups
        grid       = self.__grid
        collabels  = self.__genColumnLabels(varcols, fixedcols)
        nrows      = len(fgroups)
        ncols      = len(collabels)

        grid.ClearGrid()

        if len(fgroups) == 0:
            return

        grid.SetGridSize(nrows, ncols)
        grid.SetDragLimit(len(varcols) - 1)
        grid.SetColLabels(collabels)
        grid.ShowColLabels()
        self.__populateGrid()
        self.__createNotes()
        grid.Refresh()


    def NotesColumn(self, notes):
        """Set the position of the *Notes* column, either ``'left'`` or
        ``'right'`` of the fixed variable columns.
        """

        if notes not in ('left', 'right'):
            raise ValueError('Invalid value for notes: {}'.format(notes))

        nrows, ncols = self.__grid.GetGridSize()
        oldnotes     = self.__notes
        self.__notes = notes

        if nrows    == 0:     return
        if oldnotes == notes: return

        nvarcols = len(self.__mgr.varcols)
        order    = list(range(ncols))

        if notes == 'left': order.insert(nvarcols, order.pop(-1))
        else:               order.append(order.pop(nvarcols))

        self.__grid.ReorderColumns(order)
        self.__grid.Refresh()


    def GridContents(self):
        """Returns the contents of the grid as a list of lists of strings. """

        grid         = self.__grid
        nrows, ncols = grid.GetGridSize()
        rows         = [[None] * ncols for i in range(nrows)]

        for row, col in it.product(range(nrows), range(ncols)):

            widget = grid.GetWidget(row, col)

            # widget is either a StaticText
            # or a TextCtrl (notes column)
            if isinstance(widget, wx.TextCtrl): value = widget.GetValue()
            else:                               value = widget.GetLabel()

            if value == PRESENTLBL:
                value = 'x'

            rows[row][col] = value

        rows = [grid.GetColLabels()] + rows
        return rows


    def __populateGrid(self):
        """Populates the contents of the file tree grid. The contents
        are retrieved from the :class:`.FileTreeManager`.
        """

        mgr      = self.__mgr
        grid     = self.__grid
        fgroups  = mgr.filegroups
        varcols  = mgr.varcols
        fixedoff = 1 if self.__notes == 'left' else 0
        msg      = strings.messages[self, 'buildingList']

        fwoverlay.textOverlay(self.__grid, msg)

        for rowi, fgroup in enumerate(fgroups):

            # Varying variable columns display
            # the current variable value
            for coli, col in enumerate(varcols):
                val = fgroup.varyings[col]
                if val is None:
                    val = NONELBL
                grid.SetText(rowi, coli, val)

            # Fixed variable columns have a
            # bullet indicating whether or
            # not each file is present
            for coli, filename in enumerate(fgroup.files,
                                            len(varcols) + fixedoff):
                if filename is not None: grid.SetText(rowi, coli, PRESENTLBL)
                else:                    grid.SetText(rowi, coli, '')

        self.Refresh()
        self.Update()


    def __notesIndex(self):
        """Returns the column index of the notes column. Assumes that a
        :class:`.FileTreeManager` has been passed to :meth:`.ResetGrid`.
        """

        nvarcols   = len(self.__mgr.varcols)
        nfixedcols = len(self.__mgr.fixedcols)

        if self.__notes == 'left': return nvarcols
        else:                      return nvarcols + nfixedcols


    def __createNotes(self):
        """Called by :meth:`ResetGrid`. Creates a ``wx.TextCtrl`` for each
        row, and adds it to the end column.
        """

        grid         = self.__grid
        notesIdx     = self.__notesIndex()
        nrows, ncols = grid.GetGridSize()

        for i in range(nrows):
            note = wx.TextCtrl(grid, i)
            grid.SetWidget(i, notesIdx, note)
            note.Bind(wx.EVT_CHAR_HOOK, self.__noteCharHook)


    def __noteCharHook(self, ev):
        """Called on character events for any of the note controls. If the
        character is a tab, up, or down arrow, focus is shifted to the next
        or previous note.
        """

        key   = ev.GetKeyCode()
        shift = ev.ShiftDown()

        if key not in (wx.WXK_TAB, wx.WXK_DOWN, wx.WXK_UP):
            ev.Skip()
            return

        if (key == wx.WXK_UP) or ((key == wx.WXK_TAB) and shift): offset = -1
        else:                                                     offset = 1

        grid         = self.__grid
        notesIdx     = self.__notesIndex()
        nrows, ncols = grid.GetGridSize()
        oldrow       = ev.GetEventObject().GetId()
        newrow       = oldrow + offset

        if   newrow < 0:      newrow = 0
        elif newrow >= nrows: newrow = nrows - 1

        if newrow != oldrow:
            grid.GetWidget(newrow, notesIdx).SetFocus()
            grid.SetSelection(newrow, -1)


    def __onSelect(self, ev):
        """Called when the user selects a row. Calls the
        :meth:`.FileTreeManager.Show` method.
        """

        overlayList = self.__ftpanel.overlayList
        group       = self.__mgr.filegroups[ev.row]

        # Disable events on the grid while
        # the overlays are being swapped,
        # otherwise FSLeyes will explode.
        freezeTime = time.time()
        freezeMsg  = strings.messages[self, 'loading']

        def freeze():
            self.__grid.SetEvtHandlerEnabled(False)
            idle.idle(fwoverlay.textOverlay, self.__grid, freezeMsg)
            self.__mgr.show(group)

        def thaw():
            self.__grid.SetEvtHandlerEnabled(True)
            self.Refresh()
            self.Update()

        # Wait until every overlay is in the
        # overlay list, or 5 seconds have
        # elapsed, before re-enabling the grid.
        def thawWhen():

            thawTime = time.time()
            files    = [f for f in group.files if f is not None]
            allin    = all([overlayList.find(f) is not None for f in files])

            return allin or ((thawTime - freezeTime) >= 5)

        try:
            freeze()
        finally:
            idle.idleWhen(thaw, thawWhen)


    def __onReorder(self, ev):
        """Called when the user drags a column to change the column order.
        Calls the :meth:`.FileTreeManager.reorder` method, and updates
        the grid contents.
        """

        mgr      = self.__mgr
        grid     = self.__grid
        nvarcols = len(mgr.varcols)
        varcols  = grid.GetColLabels()[:nvarcols]

        mgr.reorder(varcols)
        self.__populateGrid()
        grid.Refresh()


    def __genColumnLabels(self, varcols, fixedcols):
        """Called by :meth:`ResetGrid`. Generates a label for each column in
        the grid, including the *Notes* column if it is visible.

        :arg varcols:   List of varying variable names

        :arg fixedcols: List of ``(name, { var : val })`` tuples, containing
                        the file type and variable values of all fixed variable
                        columns.

        :returns:       A list of labels for each column
        """

        # The first set of columns correspond
        # to the varying variables - they are
        # just labelled with the variable name.
        collabels = list(varcols)

        # The second set of columns each
        # correspond to a combination of all
        # selected file types, and all
        # selected fixed variables.
        for ftype, ftvals in fixedcols:

            # No fixed variables for this type -
            # just use the file type name
            if len(ftvals) == 0:
                lbl = ftype

            # Generate a label containing the
            # file type name, and the values
            # of all fixed variables
            else:
                lbl = ['{}={}'.format(var, val) for var, val in ftvals.items()]
                lbl = ftype + '[' + ','.join(lbl) + ']'

            collabels.append(lbl)

        # notes column (if visible) is on the
        # left or right of the fixed columns
        if self.__notes == 'left':
            collabels.insert(len(varcols), strings.labels[self, 'notes'])
        else:
            collabels.append(strings.labels[self, 'notes'])

        return collabels



#!/usr/bin/env python
#
# __init__.py - Superclasses for objects which perform 'actions'.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This package provides a collection of actions, classes - the
:class:`~fsleyes.actions.base.Action` class and the :class:`.ActionProvider`
class, and the :func:`action` and :func:`toggleAction` decorators.


The :class:`.Action` class represents some sort of action which may be
performed, enabled and disabled, and may be bound to a GUI menu item or
button. The :class:`ActionProvider` class represents some entity which can
perform one or more actions.  As the :class:`.FSLeyesPanel` class derives from
:class:`ActionProvider` pretty much everything in FSLeyes is an
:class:`ActionProvider`.


Many of the modules in this package also contain standalone functions for doing
various things, such as the :func:`.screenshot.screenshot` function, and the
:func:`.loadoverlay.loadImage` function.


The :func:`action` and :func:`toggleAction` functions are intended to be used
as decorators upon the methods of a class which derives from
:class:`ActionProvider`. For example::

    >>> import fsleyes.actions as actions
    >>> class Thing(actions.ActionProvider):
            @actions.action
            def doFirstThing(self):
                print 'First thing done'
            @actions.action
            def doSecondThing(self):
                print 'Second thing done'
            @actions.toggleAction
            def toggleOtherThing(self):
                print 'Other thing toggled'


In this example, when an instance of ``Thing`` is defined, each of the methods
that are defined as actions will be available through the methods defined in
the :class:`ActionProvder`. For example::

    >>> t = Thing()
    >>> print t.getActions()
    [('doFirstThing', Action(doFirstThing)),
     ('doSecondThing', Action(doSecondThing)),
     ('toggleOtherThing', ToggleAction(toggleOtherThing))]


You can enable/disable actions through the :meth:`ActionProvider.enableAction`
and :meth:`ActionProvider.disableAction` methods::

    >>> t.disableAction('doFirstThing')
    >>> t.doFirstThing()
    Traceback (most recent call last):
      File "<stdin>", line 1, in <module>
      File "/.../fsleyes/actions/__init__.py", line 139, in __call__
        self.__name))
    fsleyes.actions.ActionDisabledError: Action doFirstThing is disabled


It is useful to know that each method on the ``t`` instance has actually been
replaced with an :class:`.Action` instance, which encapsulates the method.
Using this knowledge, you can access the ``Action`` instances directly::

    >>> t.doFirstThing.enabled = True
    >>> t.doFirstThing()
    First thing done


The :meth:`.Action.bindToWidget` method allows a widget to be bound to an
:class:`.Action`. For example::

    # We're assuming here that a wx.App, and
    # a parent window, has been created
    >>> button = wx.Button(parent, label='Do the first thing!')
    >>> t.doSecondThing.bindToWidget(parent, button, wx.EVT_BUTTON)


All bound widgets of an ``Action`` can be accessed through the
:meth:`.Action.getBoundWidgets` method, and can be unbound via the
:meth:`.Action.unbindAllWidgets` method.


This module also provides two classes which allow a widget to be automatically
created for, and bound to an ``Action`` or ``ToggleAction`` (through the
:mod:`props.build` package):

 .. autosummary::
    :nosignatures:

    ActionButton
    ToggleActionButton
"""


import logging
import types
import inspect
import warnings
import functools

import fsleyes_props   as props
import fsleyes.strings as strings

from . import base
from . import togglecontrolpanel


Action                   = base.Action
ActionDisabledError      = base.ActionDisabledError
ToggleAction             = base.ToggleAction
ToggleControlPanelAction = togglecontrolpanel.ToggleControlPanelAction


log = logging.getLogger(__name__)


def action(*args, **kwargs):
    """A decorator which identifies a class method as an action. """
    return ActionFactory(Action, *args, **kwargs)


def toggleAction(*args, **kwargs):
    """A decorator which identifies a class method as a toggle action. """
    return ActionFactory(ToggleAction, *args, **kwargs)


def toggleControlAction(*args, **kwargs):
    """A decorator which identifies a class method as a
    :class:`.ToggleControlPanelAction`.
    """
    return ActionFactory(ToggleControlPanelAction, *args, **kwargs)


class ActionProvider(object):
    """The ``ActionProvider`` class is intended to be used as a base class for
    classes which contain actions. The :func:`action` and :func:`toggleAction`
    functions can be used as decorators on class methods, to denote them as
    actions.
    """


    def __init__(self, overlayList, displayCtx):
        """Create an ``ActionProvider``.

        :arg overlayList: The :class:`.OverlayList`.
        :arg displayCtx:  The :class:`.DisplayContext`.
        """
        self.__overlayList = overlayList
        self.__displayCtx  = displayCtx


    @property
    def overlayList(self):
        """Return a reference to the :class:`.OverlayList`. """
        return self.__overlayList


    @property
    def displayCtx(self):
        """Return a reference to the :class:`.DisplayContext`. """
        return self.__displayCtx


    def destroy(self):
        """Must be called when this ``ActionProvider`` is no longer needed.
        Calls the :meth:`Action.destroy` method of all ``Action`` instances.
        """

        self.__overlayList = None
        self.__displayCtx  = None

        for name, actionz in ActionProvider.getActions(self):

            # Entries in getActions may be (None, None)
            if actionz is None:
                continue

            # Or may be ('groupName', [('actionName', action), ...])
            elif isinstance(actionz, list):
                actionz = [a[1] for a in actionz]

            # Or may be ('actionName', action)
            else:
                actionz = [actionz]

            for a in actionz:
                a.destroy()


    def getAction(self, name):
        """Return the :class:`Action` instance with the specified name. """
        return getattr(self, name)


    def hasAction(self, name):
        """Return ``True`` if this ``ActionProvider`` has an action with the
        given name, ``False`` otherwise
        """
        return getattr(self, name, None) is not None


    def enableAction(self, name, enable=True):
        """Enable/disable the named :class:`Action`. """
        self.getAction(name).enabled = enable


    def disableAction(self, name):
        """Disable the named :class:`Action`. """
        self.enableAction(name, False)


    def getActions(self):
        """Return a list containing the ``(name, Action)`` of all
        :class:`Action` instances in this ``ActionProvider``.

        Sub-classes may wish to override this method to enforce a specific
        ordering of their actions.

        .. note:: The list returned by this method may contain:

                    - Entries equal to ``(None, None)``. This is used as a
                      hint for GUIs which display action widgets/menu items to
                      indicate that a separator should be inserted.

                    - Entries equal to ``(name, [(name, Action), ...])``.
                      This is used as a hint for GUIs that the nested list
                      of actions should appear as a sub-menu with the
                      specified ``name``.
        """

        acts = []

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            for name, attr in inspect.getmembers(self):
                if isinstance(attr, Action):
                    acts.append((name, attr))

        return acts


class ActionFactory(object):
    """The ``ActionFactory`` is used by the :func:`action` and
    :func:`toggleAction` decorators. Its job is to create :class:`Action`
    instances for :class:`ActionProvider` instances.


    .. warning:: This class contains difficult-to-understand code. Read up
                 on decorators and descriptors before proceeding.


    .. note:: This class has no use outside of this module, except for use
              with custom :class:`.Action`/:class:`.ToggleAction` sub-classes
              and corresponding decorators (along the lines of :func:`.action`
              and :func:`.toggleAction`). A custom decorator simply needs
              to return ``ActionFactory(CustomActionClass, *args, **kwargs)``,
              where the ``*args`` and ``**kwargs`` are the arguments passed to
              the :class:`Action` sub-class.

              See the :func:`toggleControlAction` decorator for an example.


    *Boring technical details*


    Consider the following class::

        class MyThing(ActionProvider):

            @action
            def myAction(self):
                # do things here


    The ``MyClass.myAction`` method has been marked as an action, using the
    :func:`action` decorator. However, the :func:`action` decorator cannot
    create an :class:`Action` instance at the point of class definition,
    because this would lead to a single ``Action`` instance being shared by
    multiple ``MyThing`` instances.  We need to be able to create an ``Action``
    instance for every ``MyThing`` instance, whilst still allowing the action
    decorator to be used on class methods.


    So when the :func:`action` or :func:`toggleAction` is used in a class
    definition, an ``ActionFactory`` is created, and used as the decorator
    of the unbound class method.


    Later on, when the ``ActionFactory`` detects that it being is accessed
    through an instance of the class (a ``MyThing`` instance in the example
    above), it creates an :class:`Action` instance, and then replaces itself
    with this ``Action`` instance - the ``Action`` instance becomes the
    decorator of the bound method. This is possible because the
    ``ActionFactory`` is a descriptor - it uses the :meth:`__get__` method
    so it can differentiate between class-level and instance-level accesses
    of the decorated method.


    The ``ActionFactory`` supports class-method decorators both with and
    without arguments. While neither the :class:`.Action`, nor the
    :class:`.ToggleAction` classes accept any optional arguments, this may be
    useful for custom sub-classes, i.e.::

        class MyThing(ActionProvider):

            @action
            def myAction(self):
                # do things here

            @customAction()
            def myAction2(self):
                # do things here

            @otherCustomAction(arg1=8)
            def myAction3(self):
                # do things here


    .. todo:: Merge/replace this class with the :class:`.memoize.Instanceify`
              decorator.
    """


    def __init__(self, actionType, *args, **kwargs):
        """Create an ``ActionFactory``.

        :arg actionType: The action type (e.g. :class:`Action` or
                         :class:`ToggleAction`).

        The remaining arguments may comprise a single callable object, (an
        ``@action`` style decorator was used) or a collection of arguments
        passed to the decorator (an ``@action(...)`` style decorator was
        used).
        """

        self.__actionType  = actionType
        self.__args        = args
        self.__kwargs      = kwargs
        self.__func        = None

        # A no-brackets style
        # decorator was used
        if len(kwargs) == 0 and \
           len(args)   == 1 and \
           isinstance(args[0], (types.FunctionType,
                                types.MethodType)):

            self.__func = args[0]
            self.__args = self.__args[1:]


    def __call__(self, func=None):
        """If this ``ActionFactory`` was instantiated through a brackets-style
        decorator (e.g. ``@action(arg1=1, arg2=2)``), this method is called
        immediately after :meth:`__init__`, with a reference to the decorated
        function. Otherwise, (an ``@action`` style decorator was used), this
        method should never be called.
        """

        if self.__func is not None:
            log.warn('ActionFactory.__call__ was called, but function is '
                     'alreday set ({})! I\'m really confused.'.format(
                         self.__func.__name__))

        self.__func = func
        return self


    def __get__(self, instance, cls):
        """When this ``ActionFactory`` is accessed through an instance,
        an :class:`Action` instance is created. This ``ActionFactory`` is
        then replaced by the ``Action`` instance.

        If this ``ActionFactory`` is accessed through a class, the
        encapsulated function is returned.
        """

        # Class-level access
        if instance is None:
            return self.__func

        else:

            # Create an Action for the instance
            action = self.__actionType(
                instance.overlayList,
                instance.displayCtx,
                self.__func,
                instance,
                *self.__args,
                **self.__kwargs)

            # and replace this ActionFactory
            # with the Action on the instance.
            setattr(instance, self.__func.__name__, action)
            return functools.update_wrapper(action, self.__func)


class ActionButton(props.Button):
    """Extends the :class:`props.Button` class to encapsulate an
    :class:`Action` instance.

    Only actions which are defined using the :func:`action` or
    :func:`toggleAction` decorator are supported.
    """


    def __init__(self,
                 actionName,
                 classType=None,
                 actionArgs=None,
                 actionKwargs=None,
                 **kwargs):
        """Create an ``ActionButton``.

        :arg actionName: Name of the action

        :arg classType:  The type which defines the action.

        :arg kwargs:     Passed to the :class:`props.Button` constructor.
        """

        if actionArgs   is None: actionArgs   = []
        if actionKwargs is None: actionKwargs = {}

        self.__name         = actionName
        self.__actionArgs   = actionArgs
        self.__actionKwargs = actionKwargs

        text = kwargs.pop('text', None)

        if text is None:
            if classType is not None:
                text = strings.actions.get((classType, actionName), actionName)
            else:
                text = actionName

        props.Button.__init__(
            self,
            actionName,
            text=text,
            callback=self.__onButton,
            setup=self.__setup,
            **kwargs)


    def __setup(self, instance, parent, widget):
        """Called when the button is created. Binds the button widget to the
        ``Action`` instance.
        """
        import wx
        instance.getAction(self.__name).bindToWidget(
            parent, wx.EVT_BUTTON, widget)


    def __onButton(self, instance, widget):
        """Called when the button is pushed. Runs the action."""
        instance.getAction(self.__name)(*self.__actionArgs,
                                        **self.__actionKwargs)


class ToggleActionButton(props.Toggle):
    """Extends the :class:`props.Toggle` class to encapsulate a
    :class:`ToggleAction` instance.

    Only actions which are defined using the :func:`action` or
    :func:`toggleAction` decorator are supported.
    """


    def __init__(self,
                 actionName,
                 icon,
                 actionArgs=None,
                 actionKwargs=None,
                 **kwargs):
        """Create a ``ToggleActionButton``.

        :arg actionName:   Name of the action

        :arg icon:         One or two icon file names to use on the button.

        :arg actionArgs:   Positional arguments to pass to the
                           :class:`.ToggleAction` when it is invoked.

        :arg actionKwargs: Keyword arguments to pass to the
                           :class:`.ToggleAction` when it is invoked.

        :arg kwargs:       Passed to the :class:`props.Toggle` constructor.
        """

        if actionArgs   is None: actionArgs   = []
        if actionKwargs is None: actionKwargs = {}

        self.__name         = actionName
        self.__actionArgs   = actionArgs
        self.__actionKwargs = actionKwargs

        props.Toggle.__init__(
            self,
            key=actionName,
            icon=icon,
            setup=self.__setup,
            callback=self.__onToggle,
            **kwargs)


    def __setup(self, instance, parent, widget):
        """Called when the toggle widget is created. Binds the widget to the
        ``ToggleAction`` instance.
        """
        import wx
        import fsleyes_widgets.bitmaptoggle as bmptoggle

        if isinstance(widget, wx.CheckBox):
            ev = wx.EVT_BUTTON
        elif isinstance(widget, wx.ToggleButton):
            ev = wx.EVT_TOGGLEBUTTON
        elif isinstance(widget, bmptoggle.BitmapToggleButton):
            ev = bmptoggle.EVT_BITMAP_TOGGLE

        else:
            raise RuntimeError(
                'Unknown widget {}'.format(type(widget).__name__))

        instance.getAction(self.__name).bindToWidget(parent, ev, widget)


    def __onToggle(self, instance, widget):
        """Called when the widget is toggled. Runs the action."""
        instance.getAction(self.__name)(*self.__actionArgs,
                                        **self.__actionKwargs)



#!/usr/bin/env python
#
# clearperspective.py - deprecated.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""Deprecated - see :mod:`.clearlayouts`. """


import fsl.utils.deprecated as deprecated

from . import clearlayouts


class ClearPerspectiveAction(clearlayouts.ClearLayoutsAction):

    @deprecated.deprecated('0.24.0', '1.0.0', 'use ClearLayoutsAction')
    def __init__(self, *args, **kwargs):
        clearlayouts.ClearLayoutsAction.__init__(self, *args, **kwargs)



#!/usr/bin/env python
#
# projectimagetosurface.py - the ProjectImageToSurfaceAction class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ProjectImageToSurfaceAction` class,
which allows data from an :class:`.Image` overlay to be projected onto
a :class:`.Mesh` overlay.
"""

import                  wx
import numpy         as np
import scipy.ndimage as ndi

import fsl.transform.affine as affine
import fsl.data.mesh        as fslmesh
import fsl.data.image       as fslimage
import fsleyes.strings      as strings
from . import                  base


class ProjectImageToSurfaceAction(base.NeedOverlayAction):
    """The ``ProjectImageToSurfaceAction`` class allows the user to project
    data from a volumetric :class:`.Image` overlay onto a :class:`.Mesh`
    overlay.

    A ``ProjectImageToSurfaceAction`` is active when the currently selected
    overlay is a :class:`.Mesh`. When executed, the user is prompted to
    select an :class:`.Image` overlay to project onto the mesh. Only images
    which overlap the bounding box of the ``Mesh`` are available as options.

    When the user selects an :class:`.Image`, the data from the image at each
    vertex in the mesh is retrieved using ``scipy.ndimage.map_coordinates``.
    This data is then added as an option on the :attr:`.MeshOpts.vertexData`
    property, and selected.
    """


    def __init__(self, overlayList, displayCtx, frame):
        """Create a ``ProjectImageToSurfaceAction``.

        :arg overlayList: The :class:`.OverlayList`
        :arg displayCtx:  The :class:`.DisplayContext`
        :arg frame:       The :class:`.ViewPanel` this action is associated
                          with.
        """
        super().__init__(overlayList,
                         displayCtx,
                         self.__projectImage,
                         fslmesh.Mesh)
        self.__frame = frame


    def __projectImage(self):
        """Run the ``ProjectImageToSurfaceAction``. """

        # TODO Allow user to load an image from file?

        # We need to use child DisplayOpts objects in
        # order to perform the projection, as the parent
        # DisplayOpts objects do not maintain information
        # about transformations between the different
        # coordinate systems. Any child DisplayContext
        # will do.
        displayCtx = self.__frame.focusedViewPanel.displayCtx
        mesh       = self.displayCtx.getSelectedOverlay()
        mopts      = displayCtx.getOpts(mesh)

        # any images which overlap with the
        # mesh bbox in the display coordinate
        # system are given as options
        images = []
        mbbox  = affine.transform(
            mesh.bounds, mopts.getTransform('mesh', 'display'))
        mbbox  = list(zip(*mbbox))

        for o in self.overlayList:

            if not isinstance(o, fslimage.Image):
                continue

            iopts = displayCtx.getOpts(o)
            ibbox = affine.axisBounds(
                o.shape[:3], iopts.getTransform('voxel', 'display'))
            ibbox = list(zip(*ibbox))

            if overlap(mbbox, ibbox):
                images.append(o)

        # can't find any images which
        # overlap with the mesh in the
        # display coordinate system
        if len(images) == 0:
            wx.MessageDialog(
                self.__frame,
                message=strings.messages[self, 'noOverlap'],
                style=(wx.ICON_EXCLAMATION | wx.OK)).ShowModal()
            return

        # ask the user what image
        # they want to project
        dlg = wx.SingleChoiceDialog(
            self.__frame,
            caption=strings.titles[self, 'dialog'],
            message=strings.labels[self, 'message'],
            style=wx.OK | wx.CANCEL,
            choices=[i.name for i in images])

        if dlg.ShowModal() != wx.ID_OK:
            return

        # Sample data from the selected image
        image = images[dlg.GetSelection()]
        vdata = projectImageDataOntoMesh(displayCtx, image, mesh)

        # add the vertex data to
        # the mesh, and select it
        key              = mopts.addVertexData(image.name, vdata)
        mopts.vertexData = key


def projectImageDataOntoMesh(displayCtx, image, mesh):
    """Samples data from ``image`` at every vertex on ``mesh``. Uses
    ``scipy.ndimage.map_coordinates``.
    """

    # transform the mesh vertices into
    # the image voxel coordinate system
    mopts   = displayCtx.getOpts(mesh)
    iopts   = displayCtx.getOpts(image)
    m2d     = mopts.getTransform('mesh', 'display')
    verts   = mesh.vertices
    verts   = iopts.transformCoords(verts, 'display', 'voxel', pre=m2d)

    # sample the image data at
    # every vertex location
    imgdata = image.data[iopts.index()]
    vdata   = ndi.map_coordinates(imgdata,
                                  verts.T,
                                  output=np.float64,
                                  cval=np.nan)

    return vdata


def overlap(bbox1, bbox2):
    """Returns ``True`` if the two bounding boxes overlap at all, ``False``
    otherwise.

    Both ``bbox1`` and ``bbox2`` are expected to be sequences of
    ``(low, high)`` tuples containing the bounds of each axis in the
    coordinate system that the bounding boxes are defined in.
    """

    def axisOverlap(lo1, hi1, lo2, hi2):
        return (hi1 >= lo2) and (lo1 <= hi2) or \
               (hi2 >= lo1) and (lo2 <= hi1)

    for (lo1, hi1), (lo2, hi2) in zip(bbox1, bbox2):
        if not axisOverlap(lo1, hi1, lo2, hi2):
            return False

    return True



#!/usr/bin/env python
#
# filtermain.py - Wrapper around fsleyes.main
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides an alternate FSLeyes entry point to
(func:`fsleyes.main.main`).

The :func:`main` function in this module calls :func:`fsleyes.main.main`, but
additionally intercepts and filters the standard out/error streams, and drops
useless warnings/errors which originate from underlying libraries.
"""


import os
import re
import sys
import queue
import logging
import warnings
import threading


# Filters may be either:
#   - a regex string, or
#   - a tuple of (regex string, total number of lines to skip)

FILTERS = [
    r'Adding duplicate image handler',
    r'Value error parsing header in AFM',
    r'Metadata\.framework \[Error\]',
    r'Class FIFinderSyncExtensionHost',
    r'_RegisterApplication()',
    r'FinderKit',
    r'CoreText note: Client requested name',
    (r'DeprecationWarning', 2),
    r'wx.NewId',
    r'ioloop.install',
    r'FutureWarning',
    r'ShimWarning',
    r'ClientToScreen',
    r'ScreenToClient',
    r'Gdk-WARNING',
    (r'Gtk-WARNING', 2),
    r'Gtk-Message',
    (r'Gtk-CRITICAL', 2),
    r'Glib-CRITICAL',
    r'GLib-GObject-WARNING',
    r'libGL error',
    r'Pango-WARNING',
    r'Xlib:  extension',
    r'CRITICAL \*\*',
    r'No matching fbConfigs',
    r'failed to load driver: swrast',
    r'\*\*\* BUG \*\*\*',
    r'In pixman',
    r'Set a breakpoint',
]


def filter_stream(stream, die, filters=None):
    """Intercept the given output stream, and filter it according to the
    filters above. The filter is run on a separate thread.

    :arg stream:  File-like to read from and filter.

    :arg die:     ``threading.Event`` object - when it is set the filter
                  thread will end gracefully.

    :arg filters: List of regular expressions to filter. If ``None``, defaults
                  to :attr:`FILTERS`.
    """

    if filters is None:
        filters = FILTERS

    # I only loosely understand how to manipulate
    # file descriptors. Useful resources:
    #
    #  - https://linuxmeerkat.wordpress.com/2011/12/02/\
    #    file-descriptors-explained/
    #  - https://stackoverflow.com/a/24277852
    #  - https://stackoverflow.com/a/17954769
    #  - https://stackoverflow.com/a/10759061

    # Redirect the stream into a pipe,
    # and filter the pipe output
    fd           = stream.fileno()
    oldfd        = os.dup(fd)
    piper, pipew = os.pipe()
    os.dup2(pipew, fd)
    os.close(pipew)

    fin  = os.fdopen(piper, 'r')
    fout = os.fdopen(oldfd, 'w')

    # Use a queue to pass lines from
    # the input stream to the output
    # stream.
    q = queue.Queue()

    # Use a Barrier to synchronise the
    # read, write, and calling threads
    alive = threading.Barrier(3)

    # The read thread runs forever,
    # just putting lines in the queue.
    def read_loop():
        alive.wait()
        while True:
            line = fin.readline()
            if line == '':
                break
            q.put(line)


    def testline(line):
        for pat in filters:

            if isinstance(pat, tuple): pat, skip = pat
            else:                      skip      = 1

            if re.search(pat, line):
                return skip
        return 0

    # The write thread runs until both
    # of the following are true:
    #
    #  - there are no lines in the queue
    #  - the die event has been set
    def write_loop():
        skip = 0
        alive.wait()
        while True:
            try:
                line = q.get(timeout=0.25)
            except queue.Empty:
                if die.is_set(): break
                else:            continue

            if skip > 0:
                skip -= 1
                continue

            skip = testline(line) - 1

            if skip < 0:
                fout.write(line)
                fout.flush()

        # Restore the original stream
        try:
            os.close(fd)
            os.close(piper)
            os.dup2(oldfd, fd)
            os.close(oldfd)
        except Exception:
            pass

    rt = threading.Thread(target=read_loop,  daemon=True)
    wt = threading.Thread(target=write_loop, daemon=True)
    rt.start()
    wt.start()

    return rt, wt, alive


def main(args=None):
    """Alternate FSLeyes entry point.

    Uses the :func:`filter_stream` function to filter the standard
    output/error streams, then calls :func:`fsleyes.main.main`.
    """

    warnings.filterwarnings('ignore', module='matplotlib')
    warnings.filterwarnings('ignore', module='xnat')
    warnings.filterwarnings('ignore', module='mpl_toolkits')
    warnings.filterwarnings('ignore', module='numpy')
    warnings.filterwarnings('ignore', module='h5py')
    warnings.filterwarnings('ignore', module='notebook')
    warnings.filterwarnings('ignore', module='trimesh')
    warnings.filterwarnings('ignore', module='h5py')
    warnings.filterwarnings('ignore', category=DeprecationWarning)
    logging.getLogger('nibabel')  .setLevel(logging.CRITICAL)
    logging.getLogger('trimesh')  .setLevel(logging.CRITICAL)
    logging.getLogger('traitlets').setLevel(logging.CRITICAL)

    die                        = threading.Event()
    rtstdout, wtstdout, oalive = filter_stream(sys.stdout, die)
    rtstderr, wtstderr, ealive = filter_stream(sys.stderr, die)

    import fsleyes.main as fm

    # wait until the filter
    # threads have started
    oalive.wait()
    ealive.wait()

    result = 1

    try:
        result = fm.main(args)
    except SystemExit as e:
        result = e.code
    finally:
        die.set()
        wtstderr.join()
        wtstdout.join()
        rtstdout.join()
        rtstderr.join()

    sys.exit(result)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
#
# giftiopts.py - The GiftiOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`GiftiOpts` class, which contains settings
for displaying a :class:`.GiftiMesh` overlay.
"""


import fsl.data.gifti as fslgifti
from . import            meshopts


class GiftiOpts(meshopts.MeshOpts):
    """The :class:`GiftiOpts` class, which contains settings for displaying
    a :class:`.GiftiMesh` overlay.

    Currently (as of FSLeyes |version|), the ``GiftiOpts`` class is identical
    to the :class:`.MeshOpts` class (from which it derives), with the exception
    that the initial value of the :attr:`.MeshOpts.coordSpace` property is
    set to ``'affine'``, and the :attr:`.MeshOpts.vertexData` property is
    populated with any GIFTI data files that are related to the surface file.
    """


    def __init__(self, overlay, *args, **kwargs):
        """Create a ``GiftiOpts`` instance.

        All arguments are passed to the :class:`.MeshOpts` constructor.
        """

        self.getProp('coordSpace').setAttribute(self, 'default', 'affine')
        self.coordSpace = 'affine'

        # Find surface files that
        # contain other vertex sets
        vertFiles  = [overlay.dataSource] + \
                     fslgifti.relatedFiles(overlay.dataSource,
                                           [fslgifti.ALLOWED_EXTENSIONS[0]])

        # Find files that contain
        # vertex data sets
        vdataFiles = [None] + fslgifti.relatedFiles(overlay.dataSource)

        self.getProp('vertexSet') .setChoices(vertFiles,  instance=self)
        self.getProp('vertexData').setChoices(vdataFiles, instance=self)

        meshopts.MeshOpts.__init__(self, overlay, *args, **kwargs)



#!/usr/bin/env python
#
# fsleyes_kernelspecmanager.py - Custom Jupyter notebook kernel spec manager.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import os.path as op
import            sys

from jupyter_client.kernelspec import KernelSpecManager

from fsl.utils.platform import platform as fslplatform


class FSLeyesKernelSpecManager(KernelSpecManager):
    """Custom jupyer ``KernelSpecManager`` which ensures that kernels can be
    instantiated from frozen versions of FSLeyes.
    """

    def get_kernel_spec(self, kernel_name):
        """If running frozen, patches the kernel launcher command line
        arguments to redirect to the frozen FSLeyes entry point (see
        :func:`fsleyes.actions.notebook.nbmain`).
        """

        spec = super(FSLeyesKernelSpecManager, self).get_kernel_spec(
            kernel_name)

        if not fslplatform.frozen:
            return spec

        exe = op.join(op.dirname(sys.executable), 'fsleyes')

        # replace [executable, '-m', 'ipykernel_launcher']
        spec.argv = [exe, 'notebook', 'kernel'] + spec.argv[3:]

        return spec



#!/usr/bin/env python
#
# meshopts.py - The MeshOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`MeshOpts` class, which defines settings
for displaying :class:`.Mesh` overlays.
"""


import itertools as it
import              logging

import numpy as np

import fsl.data.image       as fslimage
import fsl.data.mghimage    as fslmgh
import fsl.data.utils       as dutils
import fsl.transform.affine as affine
import fsl.utils.deprecated as deprecated
import fsleyes_props        as props

import fsleyes.overlay      as fsloverlay
import fsleyes.colourmaps   as fslcmaps
from . import display       as fsldisplay
from . import colourmapopts as cmapopts


log = logging.getLogger(__name__)


def genMeshColour(overlay):
    """Called by :meth:`MeshOpts.__init__`. Generates an initial colour for
    the given :class:`.Mesh` overlay.

    If the overlay file name looks like it was generated by the FSL FIRST
    segmentation tool, returns a colour from the ``freesurfercolorlut`` colour
    map. Otherwise returns a random colour.
    """
    filename        = str(overlay.dataSource)
    subcorticalCmap = fslcmaps.getLookupTable('freesurfercolorlut')

    if   'L_Thal' in filename: return subcorticalCmap.get(10).colour
    elif 'L_Caud' in filename: return subcorticalCmap.get(11).colour
    elif 'L_Puta' in filename: return subcorticalCmap.get(12).colour
    elif 'L_Pall' in filename: return subcorticalCmap.get(13).colour
    elif 'BrStem' in filename: return subcorticalCmap.get(16).colour
    elif 'L_Hipp' in filename: return subcorticalCmap.get(17).colour
    elif 'L_Amyg' in filename: return subcorticalCmap.get(18).colour
    elif 'L_Accu' in filename: return subcorticalCmap.get(26).colour
    elif 'R_Thal' in filename: return subcorticalCmap.get(49).colour
    elif 'R_Caud' in filename: return subcorticalCmap.get(50).colour
    elif 'R_Puta' in filename: return subcorticalCmap.get(51).colour
    elif 'R_Pall' in filename: return subcorticalCmap.get(52).colour
    elif 'R_Hipp' in filename: return subcorticalCmap.get(53).colour
    elif 'R_Amyg' in filename: return subcorticalCmap.get(54).colour
    elif 'R_Accu' in filename: return subcorticalCmap.get(58).colour

    return fslcmaps.randomBrightColour()


class MeshOpts(cmapopts.ColourMapOpts, fsldisplay.DisplayOpts):
    """The ``MeshOpts`` class defines settings for displaying :class:`.Mesh`
    overlays. See also the :class:`.GiftiOpts` and :class:`.FreesurferOpts`
    sub-classes.
    """


    colour = props.Colour()
    """The mesh colour. """


    outline = props.Boolean(default=False)
    """If ``True``, an outline of the mesh is shown. Otherwise a
    cross- section of the mesh is filled.
    """


    outlineWidth = props.Real(minval=0.1, maxval=10, default=2, clamped=False)
    """If :attr:`outline` is ``True``, this property defines the width of the
    outline in pixels.
    """


    showName = props.Boolean(default=False)
    """If ``True``, the mesh name is shown alongside it.

    .. note:: Not implemented yet, and maybe never will be.
    """


    discardClipped = props.Boolean(default=False)
    """Flag which controls clipping. When the mesh is coloured according to
    some data (the :attr:`vertexData` property), vertices with a data value
    outside of the clipping range are either discarded (not drawn), or
    they are still drawn, but not according to the data, rather with the
    flat :attr:`colour`.
    """


    vertexSet = props.Choice((None, ))
    """May be populated with the names of files which contain different
    vertex sets for the :class:`.Mesh` object.
    """


    vertexData = props.Choice((None, ))
    """May be populated with the names of files which contain data associated
    with each vertex in the mesh, that can be used to colour the mesh. When
    some vertex data has been succsessfully loaded, it can be accessed via
    the :meth:`getVertexData` method.
    """


    vertexDataIndex = props.Int(minval=0, maxval=0, default=0, clamped=True)
    """If :attr:`vertexData` is loaded, and has multiple data points per
    vertex (e.g. time series), this property controls the index into the
    data.
    """


    modulateData = props.Choice((None, ))
    """Populated with the same files available for the :attr:`vertexData`
    attribute. Used to apply the :attr:`.ColourMapOpts.modulateAlpha`
    setting.

    .. note:: There is currently no support for indexing into multi-
              dimensional modulate data (e.g. time points). A separate
              ``modulateDataIndex`` property may be added in the future.
    """


    refImage = props.Choice()
    """A reference :class:`.Image` instance which the mesh coordinates are
    in terms of.

    For example, if this :class:`.Mesh` represents the segmentation of
    a sub-cortical region from a T1 image, you would set the ``refImage`` to
    that T1 image.

    Any :class:`.Image` instance in the :class:`.OverlayList` may be chosen
    as the reference image.
    """


    useLut = props.Boolean(default=False)
    """If ``True``, and if some :attr:`vertexData` is loaded, the :attr:`lut`
    is used to colour vertex values instead of the :attr:`cmap` and
    :attr:`negativeCmap`.
    """


    lut = props.Choice()
    """If :attr:`useLut` is ``True``, a :class:`.LookupTable` is used to
    colour vertex data instead of the :attr:`cmap`/:attr:`negativeCmap`.
    """


    # This property is implicitly tightly-coupled to
    # the NiftiOpts.getTransform method - the choices
    # defined in this property are assumed to be valid
    # inputs to that method (with the exception of
    # ``'torig'``).
    coordSpace = props.Choice(('affine', 'pixdim', 'pixdim-flip', 'id',
                               'torig'),
                              default='pixdim-flip')
    """If :attr:`refImage` is not ``None``, this property defines the
    reference image coordinate space in which the mesh coordinates are
    defined (i.e. voxels, scaled voxels, or world coordinates).

    =============== =========================================================
    ``affine``      The mesh coordinates are defined in the reference image
                    world coordinate system.

    ``id``          The mesh coordinates are defined in the reference image
                    voxel coordinate system.

    ``pixdim``      The mesh coordinates are defined in the reference image
                    voxel coordinate system, scaled by the voxel pixdims.

    ``pixdim-flip`` The mesh coordinates are defined in the reference image
                    voxel coordinate system, scaled by the voxel pixdims. If
                    the reference image transformation matrix has a positive
                    determinant, the X axis is flipped.

    ``torig``       The mesh coordinates are defined in the Freesurfer
                    "Torig" / "vox2ras-tkr" coordnie system.
    =============== =========================================================

    The default value is ``pixdim-flip``, as this is the coordinate system
    used in the VTK sub-cortical segmentation model files output by FIRST.
    See also the :ref:`note on coordinate systems
    <volumeopts-coordinate-systems>`, and the :meth:`.NiftiOpts.getTransform`
    method.
    """


    flatShading = props.Boolean(default=False)
    """3D only. If ``True``, colours between vertices are not interpolated -
    each triangle is coloured with the colour assigned to the first vertex.
    Only has an effect when the mesh is being coloured with vertex data.
    """


    wireframe = props.Boolean(default=False)
    """3D only. If ``True``, the mesh is rendered as a wireframe. """


    def __init__(self, overlay, *args, **kwargs):
        """Create a ``MeshOpts`` instance.

        All other arguments are passed through to the :class:`.DisplayOpts`
        constructor.
        """

        # Set a default colour
        colour      = genMeshColour(overlay)
        self.colour = np.concatenate((colour, [1.0]))

        # ColourMapOpts.linkLowRanges defaults to
        # True, which is annoying for surfaces.
        self.linkLowRanges = False

        # A copy of the refImage property
        # value is kept here so, when it
        # changes, we can de-register from
        # the previous one.
        self.__oldRefImage = None

        # When the vertexData/modulateData properties
        # are changed, the data (and its min/max)
        # is loaded and stored in these
        # attributes. See the __vdataChanged
        # method.
        #
        # Keys used are 'vertex' and 'modulate'
        self.__vdata      = {}
        self.__vdataRange = {}

        nounbind = kwargs.get('nounbind', [])
        nounbind.extend(['refImage',
                         'coordSpace',
                         'vertexData',
                         'vertexSet',
                         'modulateData'])
        kwargs['nounbind'] = nounbind

        fsldisplay.DisplayOpts  .__init__(self, overlay, *args, **kwargs)
        cmapopts  .ColourMapOpts.__init__(self)

        self.__registered = self.getParent() is not None

        # Load all vertex data and vertex
        # sets on the parent opts instance
        if not self.__registered:
            self.addVertexSetOptions( overlay.vertexSets())
            self.addVertexDataOptions(overlay.vertexDataSets())

        # The master MeshOpts instance is just a
        # sync-slave, so we only need to register
        # property listeners on child instances
        else:

            self.overlayList.addListener('overlays',
                                         self.name,
                                         self.__overlayListChanged,
                                         immediate=True)

            self.addListener('refImage',
                             self.name,
                             self.__refImageChanged,
                             immediate=True)
            self.addListener('coordSpace',
                             self.name,
                             self.__coordSpaceChanged,
                             immediate=True)

            # We need to keep colour[3]
            # keeps colour[3] and Display.alpha
            # consistent w.r.t. each other (see
            # also MaskOpts)
            self.display.addListener('alpha',
                                     self.name,
                                     self.__alphaChanged,
                                     immediate=True)
            self        .addListener('colour',
                                     self.name,
                                     self.__colourChanged,
                                     immediate=True)

            self.addListener('vertexData',
                             self.name,
                             self.__vdataChanged,
                             immediate=True)
            self.addListener('modulateData',
                             self.name,
                             self.__vdataChanged,
                             immediate=True)
            self.addListener('vertexSet',
                             self.name,
                             self.__vertexSetChanged,
                             immediate=True)
            overlay.register(self.name,
                             self.__overlayVerticesChanged,
                             'vertices')

            self.__overlayListChanged()
            self.__updateBounds()
            self.__refImageChanged()

        # If we have inherited values from a
        # parent instance, make sure the vertex/
        # modulate data (if set) is initialised
        if self.vertexData is not None:
            self.__vdataChanged(self.vertexData, None, None, 'vertexData')
        if self.modulateData is not None:
            self.__vdataChanged(self.modulateData, None, None, 'modulateData')

        # If a reference image has not
        # been set on the parent MeshOpts
        # instance, see  if there is a
        # suitable one in the overlay list.
        if self.refImage is None:
            self.refImage = fsloverlay.findMeshReferenceImage(
                self.overlayList, self.overlay)


    def destroy(self):
        """Removes some property listeners, and calls the
        :meth:`.DisplayOpts.destroy` method.
        """

        if self.__registered:

            self.overlayList.removeListener('overlays', self.name)
            self.display    .removeListener('alpha',    self.name)
            self            .removeListener('colour',   self.name)
            self.overlay    .deregister(self.name, 'vertices')

            for overlay in self.overlayList:

                # An error could be raised if the
                # DC has been/is being destroyed
                try:

                    display = self.displayCtx.getDisplay(overlay)
                    opts    = self.displayCtx.getOpts(   overlay)

                    display.removeListener('name', self.name)

                    if overlay is self.refImage:
                        opts.removeListener('transform', self.name)

                except Exception:
                    pass

        self.__oldRefImage = None
        self.__vdata       = None

        cmapopts  .ColourMapOpts.destroy(self)
        fsldisplay.DisplayOpts  .destroy(self)


    @classmethod
    def getVolumeProps(cls):
        """Overrides :meth:`DisplayOpts.getVolumeProps`. Returns a list
        of property names which control the displayed volume/timepoint.
        """
        return ['vertexDataIndex']


    def getDataRange(self):
        """Overrides the :meth:`.ColourMapOpts.getDisplayRange` method.
        Returns the display range of the currently selected
        :attr:`vertexData`, or ``(0, 1)`` if none is selected.
        """
        vdata = self.__vdataRange.get('vertex')
        if vdata is None: return (0, 1)
        else:             return vdata


    def getModulateRange(self):
        """Overrides the :meth:`.ColourMapOpts.getModulateRange` method.
        Returns the display range of the currently selected
        :attr:`vertexData`, or ``None`` if none is selected.
        """
        return self.__vdataRange.get('modulate')


    def getVertexData(self, vdtype='vertex'):
        """Returns the :attr:`.MeshOpts.vertexData` or :attr:`modulateData` ,
        if some is loaded.  Returns ``None`` otherwise.
        """
        return self.__vdata.get(vdtype)


    def vertexDataLen(self):
        """Returns the length (number of data points per vertex) of the
        currently selected :attr:`vertexData`, or ``0`` if no vertex data is
        selected.
        """

        vdata = self.__vdata.get('vertex')

        if vdata is None:
            return 0

        elif len(vdata.shape) == 1:
            return 1

        else:
            return vdata.shape[1]


    def addVertexData(self, key, data):
        """Adds the given data as a  vertex data set to the :class:`.Mesh`
        overlay associated with this ``MeshOpts``.

        :arg key:  A unique key to identify the data. If a vertex data set
                   with the key already exists, a unique one is generated
                   and returned.

        :arg data: ``numpy`` array containing per-vertex data.

        :returns:  The key used to identify the data (typically equal to
                   ``key``)
        """
        count   = 1
        origKey = key
        sets    = self.overlay.vertexDataSets()

        # generate a unique key for the
        # vertex data if one with the
        # given key already exists
        while key in sets:
            key   = '{} [{}]'.format(origKey, count)
            count = count + 1

        self.overlay.addVertexData(key, data)
        self.addVertexDataOptions([key])

        return key


    def addVertexDataOptions(self, paths):
        """Adds the given sequence of paths as options to the
        :attr:`vertexData` property. It is assumed that the paths refer
        to valid vertex data files for the overlay associated with this
        ``MeshOpts`` instance.
        """

        vdataProp = self.getProp('vertexData')
        mdataProp = self.getProp('modulateData')
        newPaths  = paths
        paths     = vdataProp.getChoices(instance=self)
        paths     = paths + [p for p in newPaths if p not in paths]

        vdataProp.setChoices(paths, instance=self)
        mdataProp.setChoices(paths, instance=self)


    def addVertexSetOptions(self, paths):
        """Adds the given sequence of paths as options to the
        :attr:`vertexSet` property. It is assumed that the paths refer
        to valid vertex files for the overlay associated with this
        ``MeshOpts`` instance.
        """

        vsetProp = self.getProp('vertexSet')
        newPaths = paths
        paths    = vsetProp.getChoices(instance=self)
        paths    = paths + [p for p in newPaths if p not in paths]

        vsetProp.setChoices(paths, instance=self)


    def getConstantColour(self):
        """Returns the current :attr::`colour`, adjusted according to the
        current :attr:`.Display.brightness`, :attr:`.Display.contrast`, and
        :attr:`.Display.alpha`.
        """

        display = self.display

        # Only apply bricon if there is no vertex data assigned
        if self.vertexData is None:
            brightness = display.brightness / 100.0
            contrast   = display.contrast   / 100.0
        else:
            brightness = 0.5
            contrast   = 0.5

        colour = list(fslcmaps.applyBricon(
            self.colour[:3], brightness, contrast))

        colour.append(display.alpha / 100.0)

        return colour


    @property
    def referenceImage(self):
        """Overrides :meth:`.DisplayOpts.referenceImage`.

        If a :attr:`refImage` is selected, it is returned. Otherwise,``None``
        is returned.
        """
        return self.refImage


    @deprecated.deprecated('0.22.3', '1.0.0', 'Use getTransform instead')
    def getCoordSpaceTransform(self):
        """Returns a transformation matrix which can be used to transform
        the :class:`.Mesh` vertex coordinates into the display
        coordinate system.

        If no :attr:`refImage` is selected, this method returns an identity
        transformation.
        """

        if self.refImage is None or self.refImage not in self.overlayList:
            return np.eye(4)

        opts = self.displayCtx.getOpts(self.refImage)

        return opts.getTransform(self.coordSpace, opts.transform)


    def getVertex(self, xyz=None, tol=1):
        """Returns an integer identifying the index of the mesh vertex that
        coresponds to the given ``xyz`` location, assumed to be specified
        in the display coordinate system.

        :arg xyz: Location to convert to a vertex index. If not provided, the
                  current :class:`.DisplayContext.location` is used.

        :arg tol: Tolerance in mesh coordinate system units - if ``xyz`` is
                  farther than ``tol`` to any vertex, ``None`` is returned.
                  Pass in ``None`` to always return the nearest vertex.
        """

        if xyz is None:
            xyz = self.displayCtx.location

        xyz        = self.transformCoords(xyz, 'display', 'mesh')
        xyz        = np.asarray(xyz).reshape(1, 3)
        dist, vidx = self.overlay.trimesh.nearest.vertex(xyz)
        dist       = dist[0]
        vidx       = vidx[0]

        if tol is not None and dist > tol: return None
        else:                              return vidx


    def normaliseSpace(self, space):
        """Used by :meth:`transformCoords` and :meth:`getTransform` to
        normalise their ``from_`` and ``to`` parameters.
        """
        if space not in ('world', 'display', 'mesh', 'voxel', 'id'):
            raise ValueError('Invalid space: {}'.format(space))

        if space == 'mesh':  space = self.coordSpace
        if space == 'torig': space = 'affine'

        return space


    def transformCoords(self, coords, from_, to, *args, **kwargs):
        """Transforms the given ``coords`` from ``from_`` to ``to``.

        :arg coords: Coordinates to transform.
        :arg from_:  Space that the coordinates are in
        :arg to:     Space to transform the coordinates to

        All other parameters are passed through to the
        :meth:`.NiftiOpts.transformCoords` method of the reference image
        ``DisplayOpts``.

        The following values are accepted for the ``from_`` and ``to``
        parameters:

          - ``'world'``:  World coordinate system
          - ``'display'`` Display coordinate system
          - ``'mesh'``    The coordinate system of this mesh.
          - ``'voxel'``:   The voxel coordinate system of the reference
                           image
          - ``'id'``:      Equivalent to ``'voxel'``.
        """

        nfrom_ = self.normaliseSpace(from_)
        nto    = self.normaliseSpace(to)
        ref    = self.refImage
        pre    = None
        post   = None

        if ref is None:
            return coords

        if from_ == 'mesh' and self.coordSpace == 'torig':
            pre = affine.concat(ref.getAffine('voxel', 'world'),
                                affine.invert(fslmgh.voxToSurfMat(ref)))

        if to == 'mesh' and self.coordSpace == 'torig':
            post = affine.concat(fslmgh.voxToSurfMat(ref),
                                 ref.getAffine('world', 'voxel'))

        opts = self.displayCtx.getOpts(ref)

        return opts.transformCoords(
            coords, nfrom_, nto, pre=pre, post=post, **kwargs)


    def getTransform(self, from_, to):
        """Return a matrix which may be used to transform coordinates from
        ``from_`` to ``to``.

        If the :attr:`refImage` property is not set, an identity matrix is
        returned.

        The following values are accepted for the ``from_`` and ``to``
        parameters:

          - ``'world'``:   World coordinate system
          - ``'display'``: Display coordinate system
          - ``'mesh'``:    The coordinate system of this mesh.
          - ``'voxel'``:   The voxel coordinate system of the reference
                           image
          - ``'id'``:      Equivalent to ``'voxel'``.
        """

        nfrom_ = self.normaliseSpace(from_)
        nto    = self.normaliseSpace(to)
        ref    = self.refImage

        if ref is None:
            return np.eye(4)

        opts  = self.displayCtx.getOpts(ref)
        xform = opts.getTransform(nfrom_, nto)

        if from_ == 'mesh' and self.coordSpace == 'torig':
            surfToVox = affine.invert(fslmgh.voxToSurfMat(ref))
            xform     = affine.concat(xform,
                                      ref.getAffine('voxel', 'world'),
                                      surfToVox)
        if to == 'mesh' and self.coordSpace == 'torig':
            voxToSurf = fslmgh.voxToSurfMat(ref)
            xform     = affine.concat(voxToSurf,
                                      ref.getAffine('world', 'voxel'),
                                      xform)

        return xform


    def __transformChanged(self, value, valid, ctx, name):
        """Called when the :attr:`.NiftiOpts.transform` property of the current
        :attr:`refImage` changes. Calls :meth:`__updateBounds`.
        """
        self.__updateBounds()


    def __coordSpaceChanged(self, *a):
        """Called when the :attr:`coordSpace` property changes.
        Calls :meth:`__updateBounds`.
        """
        self.__updateBounds()


    def __refImageChanged(self, *a):
        """Called when the :attr:`refImage` property changes.

        If a new reference image has been specified, removes listeners from
        the old one (if necessary), and adds listeners to the
        :attr:`.NiftiOpts.transform` property associated with the new image.
        Calls :meth:`__updateBounds`.
        """

        # TODO You are not tracking changes to the
        # refImage overlay type -  if this changes,
        # you will need to re-bind to the transform
        # property of the new DisplayOpts instance

        if self.__oldRefImage is not None and \
           self.__oldRefImage in self.overlayList:

            opts = self.displayCtx.getOpts(self.__oldRefImage)
            opts.removeListener('transform', self.name)

        self.__oldRefImage = self.refImage

        if self.refImage is not None:
            opts = self.displayCtx.getOpts(self.refImage)
            opts.addListener('transform',
                             self.name,
                             self.__transformChanged,
                             immediate=True)

        self.__updateBounds()


    def __updateBounds(self):
        """Called whenever any of the :attr:`refImage`, :attr:`coordSpace`,
        or :attr:`transform` properties change.

        Updates the :attr:`.DisplayOpts.bounds` property accordingly.
        """

        # create a bounding box for the
        # overlay vertices in their
        # native coordinate system
        lo, hi        = self.overlay.bounds
        xlo, ylo, zlo = lo
        xhi, yhi, zhi = hi

        # Transform the bounding box
        # into display coordinates
        xform         = self.getTransform('mesh', 'display')
        bbox          = list(it.product(*zip(lo, hi)))
        bbox          = affine.transform(bbox, xform)

        # re-calculate the min/max bounds
        x        = np.sort(bbox[:, 0])
        y        = np.sort(bbox[:, 1])
        z        = np.sort(bbox[:, 2])
        xlo, xhi = x.min(), x.max()
        ylo, yhi = y.min(), y.max()
        zlo, zhi = z.min(), z.max()

        oldBounds   = self.bounds
        self.bounds = [xlo, xhi, ylo, yhi, zlo, zhi]

        if np.all(np.isclose(oldBounds, self.bounds)):
            self.propNotify('bounds')


    def __overlayListChanged(self, *a):
        """Called when the overlay list changes. Updates the :attr:`refImage`
        property so that it contains a list of overlays which can be
        associated with the mesh.
        """

        imgProp  = self.getProp('refImage')
        imgVal   = self.refImage
        overlays = self.displayCtx.getOrderedOverlays()

        # the overlay for this MeshOpts
        # instance has been removed
        if self.overlay not in overlays:
            self.overlayList.removeListener('overlays', self.name)
            return

        imgOptions = [None]

        for overlay in overlays:

            # The overlay must be a Nifti instance.
            if not isinstance(overlay, fslimage.Nifti):
                continue

            imgOptions.append(overlay)

            display = self.displayCtx.getDisplay(overlay)
            display.addListener('name',
                                self.name,
                                self.__overlayListChanged,
                                overwrite=True)

        # The previous refImage may have
        # been removed from the overlay list
        if imgVal in imgOptions: self.refImage = imgVal
        else:                    self.refImage = None

        imgProp.setChoices(imgOptions, instance=self)


    def __overlayVerticesChanged(self, *a):
        """Called when the :attr:`.Mesh.vertices` change. Makes sure that the
        :attr:`vertexSet` attribute is synchronised.
        """

        vset   = self.overlay.selectedVertices()
        vsprop = self.getProp('vertexSet')

        if vset not in vsprop.getChoices(instance=self):
            self.addVertexSetOptions([vset])
        self.vertexSet = vset


    def __vertexSetChanged(self, *a):
        """Called when the :attr:`.MeshOpts.vertexSet` property changes.
        Updates the current vertex set on the :class:`.Mesh` overlay, and
        the overlay bounds.
        """

        if self.vertexSet not in self.overlay.vertexSets():
            self.overlay.loadVertices(self.vertexSet)
        else:
            with self.overlay.skip(self.name, 'vertices'):
                self.overlay.vertices = self.vertexSet

        self.__updateBounds()


    def __vdataChanged(self, value, valid, ctx, name):
        """Called when the :attr:`vertexData` or :attr:`modulateData`
        properties changes. Attempts to load the data if possible. The data may
        subsequently be retrieved via the :meth:`getVertexData` method.
        """

        vdata      = None
        vdataRange = None
        overlay    = self.overlay
        vdfile     = value

        if   name == 'vertexData':   key = 'vertex'
        elif name == 'modulateData': key = 'modulate'
        else: raise RuntimeError()

        try:
            if vdfile is not None:

                if vdfile not in overlay.vertexDataSets():
                    log.debug('Loading vertex data: {}'.format(vdfile))
                    vdata = overlay.loadVertexData(vdfile)
                else:
                    vdata = overlay.getVertexData(vdfile)

                vdataRange = np.nanmin(vdata), np.nanmax(vdata)

                if len(vdata.shape) == 1:
                    vdata = vdata.reshape(-1, 1)

                vdata = dutils.makeWriteable(vdata)

        except Exception as e:

            # TODO show a warning
            log.warning('Unable to load vertex data from {}: {}'.format(
                vdfile, e, exc_info=True))

            vdata      = None
            vdataRange = None

        self.__vdata[key]      = vdata
        self.__vdataRange[key] = vdataRange

        if key == 'vertex':
            if vdata is not None: npoints = vdata.shape[1]
            else:                 npoints = 1
            self.vertexDataIndex = 0
            self.setAttribute('vertexDataIndex', 'maxval', npoints - 1)

        # if modulate data has changed,
        # don't update display/clipping
        # ranges (unless modulateData is
        # None, meaning that it is using
        # vertexData)
        if key == 'vertex':
            drange = True
            mrange = self.modulateData is None
        # and vice versa
        else:
            drange = False
            mrange = True

        self.updateDataRange(drange, drange, mrange)


    def __colourChanged(self, *a):
        """Called when :attr:`.colour` changes. Updates :attr:`.Display.alpha`
        from the alpha component.
        """

        # modulateAlpha may cause the
        # alpha property to be disabled
        if not self.display.propertyIsEnabled('alpha'):
            return

        alpha = self.colour[3] * 100

        log.debug('Propagating MeshOpts.colour[3] to '
                  'Display.alpha [{}]'.format(alpha))

        with props.skip(self.display, 'alpha', self.name):
            self.display.alpha = alpha


    def __alphaChanged(self, *a):
        """Called when :attr:`.Display.alpha` changes. Updates the alpha
        component of :attr:`.colour`.
        """

        alpha      = self.display.alpha / 100.0
        r, g, b, _ = self.colour

        log.debug('Propagating Display.alpha to MeshOpts.'
                  'colour[3] [{}]'.format(alpha))

        with props.skip(self, 'colour', self.name):
            self.colour = r, g, b, alpha



#!/usr/bin/env python
#
# orthotoolbar.py - The OrthoToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`OrthoToolBar` class, which is a
:class:`.ControlToolBar` for use with the :class:`.OrthoPanel`.
"""


import wx

import fsleyes_props    as props

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.toolbar               as fsltoolbar
import fsleyes.icons                 as fslicons
import fsleyes.tooltips              as fsltooltips
import fsleyes.actions               as actions
import fsleyes.strings               as strings


class OrthoToolBar(ctrlpanel.ControlToolBar):
    """The ``OrthoToolBar`` is a :class:`.ControlToolBar` for use with the
    :class:`.OrthoPanel`. An ``OrthoToolBar`` looks something like this:


    .. image:: images/orthotoolbar.png
       :scale: 50%
       :align: center


    The ``OrthoToolBar`` allows the user to control important parts of the
    :class:`.OrthoPanel` display, and also to display a
    :class:`.CanvasSettingsPanel`, which allows control over all aspects of
    an ``OrthoPanel``.

    The ``OrthoToolBar`` contains controls which modify properties, or run
    actions, defined on the following classes:

    .. autosummary::
       :nosignatures:

       ~fsleyes.views.orthopanel.OrthoPanel
       ~fsleyes.displaycontext.orthoopts.OrthoOpts
       ~fsleyes.profiles.orthoviewprofile.OrthoViewProfile
    """


    showCursorAndLabels = props.Boolean(default=True)
    """This property is linked to a button on the toolbar which allows the
    user to simultaneously toggle the :attr:`.SceneOpts.showCursor` and
    :attr:`.OrthoOpts.showLabels` properties.
    """


    def __init__(self, parent, overlayList, displayCtx, frame, ortho):
        """Create an ``OrthoToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg ortho:       The :class:`.OrthoPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          frame,
                                          height=24,
                                          kbFocus=True)

        self.orthoPanel = ortho

        # The toolbar has buttons bound to some actions
        # on the Profile  instance - when the profile
        # changes (between 'view' and 'edit'), the
        # Profile instance changes too, so we need
        # to re-create these action buttons. I'm being
        # lazy and just re-generating the entire toolbar.
        ortho.addListener('profile', self.name, self.__makeTools)

        self.addListener('showCursorAndLabels',
                         self.name,
                         self.__showCursorAndLabelsChanged)

        self.__makeTools()


    def destroy(self):
        """Must be called when this ``OrthoToolBar`` is no longer in use.
        Removes some property listeners, and calls the base class
        implementation.
        """
        self.orthoPanel.removeListener('profile',             self.name)
        self           .removeListener('showCursorAndLabels', self.name)

        ctrlpanel.ControlToolBar.destroy(self)


    @staticmethod
    def supportedViews():
        """Overrides :meth:`.ControlMixin.supportedViews`. The
        ``OrthoToolBar`` is only intended to be added to
        :class:`.OrthoPanel` views.
        """
        from fsleyes.views.orthopanel import OrthoPanel
        return [OrthoPanel]


    def __makeTools(self, *a):
        """Called by :meth:`__init__`, and whenever the
        :attr:`.ViewPanel.profile` property changes.

        Re-creates all tools shown on this ``OrthoToolBar``.
        """

        ortho     = self.orthoPanel
        orthoOpts = ortho.sceneOpts
        profile   = ortho.getCurrentProfile()

        coronalIcon          = 'coronalSlice24'
        coronalHighlightIcon = 'coronalSliceHighlight24'

        icons = {
            'screenshot'       : fslicons.findImageFile('camera24'),
            'resetDisplay'     : fslicons.findImageFile('resetZoom24'),
            'showCursorAndLabels' : [
                fslicons.findImageFile('addHighlight24'),
                fslicons.findImageFile('add24')],
            'movieMode'        : [
                fslicons.findImageFile('movieHighlight24'),
                fslicons.findImageFile('movie24')],
            'showXCanvas'      : [
                fslicons.findImageFile('sagittalSliceHighlight24'),
                fslicons.findImageFile('sagittalSlice24')],
            'showYCanvas'      : [
                fslicons.findImageFile(coronalHighlightIcon),
                fslicons.findImageFile(coronalIcon)],
            'showZCanvas'      : [
                fslicons.findImageFile('axialSliceHighlight24'),
                fslicons.findImageFile('axialSlice24')],
            'toggleCanvasSettingsPanel' : [
                fslicons.findImageFile('spannerHighlight24'),
                fslicons.findImageFile('spanner24')],

            'layout' : {
                'horizontal' : [
                    fslicons.findImageFile('horizontalLayoutHighlight24'),
                    fslicons.findImageFile('horizontalLayout24')],
                'vertical'   : [
                    fslicons.findImageFile('verticalLayoutHighlight24'),
                    fslicons.findImageFile('verticalLayout24')],
                'grid'       : [
                    fslicons.findImageFile('gridLayoutHighlight24'),
                    fslicons.findImageFile('gridLayout24')]}
        }

        tooltips = {
            'screenshot'   : fsltooltips.actions[   ortho,     'screenshot'],
            'resetDisplay' : fsltooltips.actions[   profile,   'resetDisplay'],
            'movieMode'    : fsltooltips.properties[ortho,     'movieMode'],
            'showCursorAndLabels' : fsltooltips.properties[
                self, 'showCursorAndLabels'],
            'zoom'         : fsltooltips.properties[orthoOpts, 'zoom'],
            'layout'       : fsltooltips.properties[orthoOpts, 'layout'],
            'showXCanvas'  : fsltooltips.properties[orthoOpts, 'showXCanvas'],
            'showYCanvas'  : fsltooltips.properties[orthoOpts, 'showYCanvas'],
            'showZCanvas'  : fsltooltips.properties[orthoOpts, 'showZCanvas'],
            'toggleCanvasSettingsPanel' : fsltooltips.actions[
                ortho, 'toggleCanvasSettingsPanel'],

        }

        targets    = {'screenshot'                : ortho,
                      'movieMode'                 : ortho,
                      'showCursorAndLabels'       : self,
                      'resetDisplay'              : profile,
                      'zoom'                      : orthoOpts,
                      'layout'                    : orthoOpts,
                      'showXCanvas'               : orthoOpts,
                      'showYCanvas'               : orthoOpts,
                      'showZCanvas'               : orthoOpts,
                      'toggleCanvasSettingsPanel' : ortho}


        toolSpecs = [

            actions.ToggleActionButton(
                'toggleCanvasSettingsPanel',
                actionKwargs={'floatPane' : True},
                icon=icons['toggleCanvasSettingsPanel'],
                tooltip=tooltips['toggleCanvasSettingsPanel']),
            actions.ActionButton('screenshot',
                                 icon=icons['screenshot'],
                                 tooltip=tooltips['screenshot']),
            'div',
            props  .Widget(      'showXCanvas',
                                 icon=icons['showXCanvas'],
                                 tooltip=tooltips['showXCanvas']),
            props  .Widget(      'showYCanvas',
                                 icon=icons['showYCanvas'],
                                 tooltip=tooltips['showYCanvas']),
            props  .Widget(      'showZCanvas',
                                 icon=icons['showZCanvas'],
                                 tooltip=tooltips['showZCanvas']),
            'div',
            props  .Widget(      'layout',
                                 icons=icons['layout'],
                                 tooltip=tooltips['layout']),
            'div',
            props  .Widget(      'movieMode',
                                 icon=icons['movieMode'],
                                 tooltip=tooltips['movieMode']),
            props  .Widget(      'showCursorAndLabels',
                                 icon=icons['showCursorAndLabels'],
                                 tooltip=tooltips['showCursorAndLabels']),
            actions.ActionButton('resetDisplay',
                                 icon=icons['resetDisplay'],
                                 tooltip=tooltips['resetDisplay']),
            props.Widget(        'zoom',
                                 spin=True,
                                 slider=True,
                                 showLimits=False,
                                 spinWidth=5,
                                 tooltip=tooltips['zoom']),
        ]

        tools = []
        nav   = []

        for spec in toolSpecs:

            if spec == 'div':
                tools.append(fsltoolbar.ToolBarDivider(self,
                                                       height=24,
                                                       orient=wx.VERTICAL))
                continue

            widget    = props.buildGUI(self, targets[spec.key], spec)
            navWidget = widget

            if spec.key in ('zoom', ):
                widget = self.MakeLabelledTool(
                    widget,
                    strings.properties[targets[spec.key], spec.key])

            tools.append(widget)
            nav  .append(navWidget)

        self.SetTools(tools, destroy=True)
        self.setNavOrder(nav)


    def __showCursorAndLabelsChanged(self, *a):
        """Called when the :attr:`showCursorAndLabels` property is changed.
        Propagates the change on to the :attr:`.SceneOpts.showCursor` and
        :attr:`.OrthoOpts.showLabels` properties.
        """

        opts            = self.orthoPanel.sceneOpts
        opts.showCursor = self.showCursorAndLabels
        opts.showLabels = self.showCursorAndLabels



#!/usr/bin/env python
#
# volume3dopts.py - The Volume3DOpts class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`.Volume3DOpts` class, a mix-in for
use with :class:`.DisplayOpts` classes.
"""

import numpy as np

from   fsl.utils.platform import platform as fslplatform
import fsl.transform.affine               as affine
import fsl.utils.deprecated               as deprecated
import fsleyes_props                      as props


class Volume3DOpts(object):
    """The ``Volume3DOpts`` class is a mix-in for use with :class:`.DisplayOpts`
    classes. It defines display properties used for ray-cast based rendering
    of :class:`.Image` overlays.


    The properties in this class are tightly coupled to the ray-casting
    implementation used by the :class:`.GLVolume` class - see its documentation
    for details.
    """


    blendFactor = props.Real(minval=0.001, maxval=1, default=0.1)
    """Controls how much each sampled point on each ray contributes to the
    final colour.
    """


    numSteps = props.Int(minval=25, maxval=500, default=100, clamped=False)
    """Specifies the maximum number of samples to acquire in the rendering of
    each pixel of the 3D scene. This corresponds to the number of iterations
    of the ray-casting loop.

    .. note:: In a low performance environment, the actual number of steps
              may differ from this value - use the :meth:`getNumSteps` method
              to get the number of steps that are actually executed.
    """


    numInnerSteps = props.Int(minval=1, maxval=100, default=10, clamped=True)
    """Only used in low performance environments. Specifies the number of
    ray-casting steps to execute in a single iteration on the GPU, as part
    of an outer loop which is running on the CPU. See the :class:`.GLVolume`
    class documentation for more details on the rendering process.

    .. warning:: The maximum number of iterations that can be performed within
                 an ARB fragment program is implementation-dependent. Too high
                 a value may result in errors or a corrupted view. See the
                 :class:`.GLVolume` class for details.
    """


    resolution = props.Int(minval=10, maxval=100, default=100, clamped=True)
    """Only used in low performance environments. Specifies the resolution
    of the off-screen buffer to which the volume is rendered, as a percentage
    of the screen resolution.

    See the :class:`.GLVolume` class documentation for more details.
    """


    smoothing = props.Int(minval=0, maxval=10, default=0, clamped=True)
    """Amount of smoothing to apply to the rendered volume - this setting
    controls the smoothing filter radius, in pixels.
    """


    numClipPlanes = props.Int(minval=0, maxval=5, default=0, clamped=True)
    """Number of active clip planes. """


    showClipPlanes = props.Boolean(default=False)
    """If ``True``, wirframes depicting the active clipping planes will
    be drawn.
    """

    clipMode = props.Choice(('intersection', 'union', 'complement'))
    """This setting controls how the active clip planes are combined.

      -  ``intersection`` clips the intersection of all planes
      -  ``union`` clips the union of all planes
      -  ``complement`` clips the complement of all planes
    """


    clipPosition = props.List(
        props.Percentage(minval=0, maxval=100, clamped=True),
        minlen=10,
        maxlen=10)
    """Centre of clip-plane rotation, as a distance from the volume centre -
    0.5 is centre.
    """


    clipAzimuth = props.List(
        props.Real(minval=-180, maxval=180, clamped=True),
        minlen=10,
        maxlen=10)
    """Rotation (degrees) of the clip plane about the Z axis, in the display
    coordinate system.
    """


    clipInclination = props.List(
        props.Real(minval=-180, maxval=180, clamped=True),
        minlen=10,
        maxlen=10)
    """Rotation (degrees) of the clip plane about the Y axis in the display
    coordinate system.
    """


    def __init__(self):
        """Create a :class:`Volume3DOpts` instance.
        """

        # If we're in an X11/SSh session,
        # step down the quality so it's
        # a bit faster.
        if fslplatform.inSSHSession:
            self.numSteps    = 60
            self.resolution  = 70
            self.blendFactor = 0.3

        # If we're in GL14, restrict the
        # maximum possible amount of
        # smoothing, as GL14 fragment
        # programs cannot be too large.
        if float(fslplatform.glVersion) < 2.1:
            smooth = self.getProp('smoothing')
            smooth.setAttribute(self, 'maxval', 6)

        self.clipPosition[:]    = 10 * [50]
        self.clipAzimuth[:]     = 10 * [0]
        self.clipInclination[:] = 10 * [0]

        # Give convenient initial values for
        # the first three clipping planes
        self.clipInclination[1] = 90
        self.clipAzimuth[    1] = 0
        self.clipInclination[2] = 90
        self.clipAzimuth[    2] = 90


    def destroy(self):
        """Does nothing. """
        pass


    @property
    @deprecated.deprecated(
        '0.17.0', '1.0.0', 'Dithering is automatically calculated')
    def dithering(self):
        """Deprecated."""
        pass


    def getNumSteps(self):
        """Return the value of the :attr:`numSteps` property, possibly
        adjusted according to the the :attr:`numInnerSteps` property. The
        result of this method should be used instead of the value of
        the :attr:`numSteps` property.

        See the :class:`.GLVolume` class for more details.
        """

        if float(fslplatform.glVersion) >= 2.1:
            return self.numSteps

        outer = self.getNumOuterSteps()

        return int(outer * self.numInnerSteps)


    def getNumOuterSteps(self):
        """Returns the number of iterations for the outer ray-casting loop.

        See the :class:`.GLVolume` class for more details.
        """

        total = self.numSteps
        inner = self.numInnerSteps
        outer = np.ceil(total / float(inner))

        return int(outer)


    def calculateRayCastSettings(self, view=None, proj=None):
        """Calculates various parameters required for 3D ray-cast rendering
        (see the :class:`.GLVolume` class).


        :arg view: Transformation matrix which transforms from model
                   coordinates to view coordinates (i.e. the GL view matrix).


        :arg proj: Transformation matrix which transforms from view coordinates
                   to normalised device coordinates (i.e. the GL projection
                   matrix).

        Returns a tuple containing:

          - A vector defining the amount by which to move along a ray in a
            single iteration of the ray-casting algorithm. This can be added
            directly to the volume texture coordinates.

          - A transformation matrix which transforms from image texture
            coordinates into the display coordinate system.

        .. note:: This method will raise an error if called on a
                  ``GLImageObject`` which is managing an overlay that is not
                  associated with a :class:`.Volume3DOpts` instance.
        """

        if view is None: view = np.eye(4)
        if proj is None: proj = np.eye(4)

        # In GL, the camera position
        # is initially pointing in
        # the -z direction.
        eye    = [0, 0, -1]
        target = [0, 0,  1]

        # We take this initial camera
        # configuration, and transform
        # it by the inverse modelview
        # matrix
        t2dmat = self.getTransform('texture', 'display')
        xform  = affine.concat(view, t2dmat)
        ixform = affine.invert(xform)

        eye    = affine.transform(eye,    ixform, vector=True)
        target = affine.transform(target, ixform, vector=True)

        # Direction that the 'camera' is
        # pointing, normalied to unit length
        cdir = affine.normalise(eye - target)

        # Calculate the length of one step
        # along the camera direction in a
        # single iteration of the ray-cast
        # loop. Multiply by sqrt(3) so that
        # the maximum number of steps will
        # be reached across the longest axis
        # of the image texture cube.
        rayStep = np.sqrt(3) * cdir / self.getNumSteps()

        # A transformation matrix which can
        # transform image texture coordinates
        # into the corresponding screen
        # (normalised device) coordinates.
        # This allows the fragment shader to
        # convert an image texture coordinate
        # into a relative depth value.
        #
        # The projection matrix puts depth into
        # [-1, 1], but we want it in [0, 1]
        zscale = affine.scaleOffsetXform([1, 1, 0.5], [0, 0, 0.5])
        xform  = affine.concat(zscale, proj, xform)

        return rayStep, xform


    def get3DClipPlane(self, planeIdx):
        """A convenience method which calculates a point-vector description
        of the specified clipping plane. ``planeIdx`` is an index into the
        :attr:`clipPosition`, :attr:`clipAzimuth`, and
        :attr:`clipInclination`, properties.

        Returns the clip plane at the given ``planeIdx`` as an origin and
        normal vector, in the display coordinate system..
        """

        pos     = self.clipPosition[   planeIdx]
        azimuth = self.clipAzimuth[    planeIdx]
        incline = self.clipInclination[planeIdx]

        b       = self.bounds
        pos     = pos             / 100.0
        azimuth = azimuth * np.pi / 180.0
        incline = incline * np.pi / 180.0

        xmid = b.xlo + 0.5 * b.xlen
        ymid = b.ylo + 0.5 * b.ylen
        zmid = b.zlo + 0.5 * b.zlen

        centre = [xmid, ymid, zmid]
        normal = [0, 0, -1]

        rot1     = affine.axisAnglesToRotMat(incline, 0, 0)
        rot2     = affine.axisAnglesToRotMat(0, 0, azimuth)
        rotation = affine.concat(rot2, rot1)

        normal = affine.transformNormal(normal, rotation)
        normal = affine.normalise(normal)

        offset = (pos - 0.5) * max((b.xlen, b.ylen, b.zlen))
        origin = centre + normal * offset

        return origin, normal



#!/usr/bin/env python
#
# filter.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`Filter` class, which provides an interface
to loading and running simple filter shader programs, which require a
:class:`.Texture2D` as their input.
"""


import collections


import OpenGL.GL                          as gl

from   fsl.utils.platform import platform as fslplatform
import fsleyes.gl.shaders                 as shaders


GL14_CONSTANTS = collections.defaultdict(dict, {
    'smooth' : ['kernSize']
})
"""This dictionary contains the names of any constant parameters that are
required by GL14 filter implementations. It is used by the :meth:`Filter.set`
method.

The :meth:`Filter.set` method allows both uniform and constant paramters to be
updated, but when a constant parameter is updated, the shader needs to be
recompiled.
"""


class Filter(object):
    """A ``Filter`` object encapsulates a shader program which applies some
    sort of image filter to a :class:`.Texture2D`.

    All filters use the same vertex shader, which is called
    ``filter_vert.glsl`` or ``filter_vert.prog``.

    Filter fragment shaders are assumed to have the name
    ``filter_[name]_frag.glsl`` or ``filter_[name]_frag.prog``, where
    ``[name]`` is the name of the filter that can be passed to
    :meth:`__init__`.

    Filter fragment shaders must define the following varyings/uniforms/
    attributes:

      - ``texture``      - the 2D texture which contains the filter input
      - ``fragTexCoord`` - The texture coordinate, as passed through from the
                           vertex shader.

    Other settable filter parameters must be declared as uniforms, and can be
    set via the :meth:`set` method.
    """


    def __init__(self, filterName, texture):
        """Create a ``Filter``.

        :arg filterName: Name of the filter to create.

        :arg texture:    Number of the texture unit that the filter input
                         texture will be bound to. This must be specified
                         when the shader program is compiled, to support
                         OpenGL 1.4.
        """

        basename        = filterName
        filterName      = 'filter_{}'.format(filterName)
        vertSrc         = shaders.getVertexShader( 'filter')
        fragSrc         = shaders.getFragmentShader(filterName)
        self.__texture  = texture
        self.__basename = basename

        if float(fslplatform.glVersion) >= 2.1:
            self.__shader = shaders.GLSLShader(vertSrc, fragSrc)
        else:
            constants = {n : 1 for n in GL14_CONSTANTS[basename]}
            self.__shader = shaders.ARBPShader(
                vertSrc,
                fragSrc,
                shaders.getShaderDir(),
                {'texture' : texture},
                constants=constants)


    def destroy(self):
        """Must be called when this ``Filter`` is no longer needed. Destroys
        the shader program.
        """
        self.__shader.destroy()
        self.__shader = None


    def set(self, **kwargs):
        """Set filter parameters. This method must be called before
        :meth:`apply` or :meth:`osApply` can be called, even if no parameters
        need to be set.

        The filter parameters vary depending on the specific filter that is
        used.
        """

        shader   = self.__shader
        texture  = self.__texture
        basename = self.__basename

        shader.load()

        kwargs        = dict(kwargs)
        glver         = float(fslplatform.glVersion)
        needRecompile = False

        if glver >= 2.1:
            kwargs['texture'] = texture

        for name, value in kwargs.items():
            if glver >= 2.1:
                shader.set(name, value)
            else:
                if name in GL14_CONSTANTS[basename]:
                    needRecompile = (needRecompile or
                                     shader.setConstant(name, value))
                else:
                    shader.setFragParam(name, value)

        if needRecompile:
            shader.recompile()

        shader.unload()


    def apply(self,
              source,
              zpos,
              xmin,
              xmax,
              ymin,
              ymax,
              xax,
              yax,
              xform=None,
              **kwargs):
        """Apply the filter to the given ``source`` texture, and render the
        results according to the given bounds.

        :arg source:      :class:`.Texture2D` instance to apply the filter to
        :arg zpos:        Position along the Z axis, in the display coordinate
                          system.
        :arg xmin:        Minimum X axis coordinate.
        :arg xmax:        Maximum X axis coordinate.
        :arg ymin:        Minimum Y axis coordinate.
        :arg ymax:        Maximum Y axis coordinate.
        :arg xax:         Display space axis which maps to the horizontal
                          screen axis.
        :arg yax:         Display space axis which maps to the vertical screen
                          axis.
        :arg xform:       Transformation matrix to appply to vertices.

        All other keyword arguments are passed through to the
        :meth:`.Texture2D.draw` method of the ``source`` texture.
        """

        shader    = self.__shader
        vertices  = source.generateVertices(
            zpos, xmin, xmax, ymin, ymax, xax, yax, xform)
        texCoords = source.generateTextureCoords()

        shader.load()
        shader.loadAtts()
        shader.setAtt('texCoord', texCoords)

        if float(fslplatform.glVersion) >= 2.1:
            shader.setAtt('vertex', vertices)
            source.draw(**kwargs)
        else:
            source.draw(vertices=vertices, **kwargs)

        shader.unloadAtts()
        shader.unload()


    def osApply(self, source, dest, clearDest=True, **kwargs):
        """Apply the filter to the given ``source`` texture, rendering
        the results to the given ``dest`` texture.

        This method can be used for ping-ponging, by using two
        :class:`.RenderTexture` objects, and swapping the ``source`` and
        ``dest`` parameters on each iteration.

        :arg source:    :class:`.Texture2D` instance to apply the filter to
        :arg dest:      :class:`.RenderTexture` instance to render the result
                        to
        :arg clearDest: If ``True`` (the default), the ``dest`` texture is
                        cleared before the draw.

        All other arguments are passed to the :meth:`apply` method.
        """

        with dest.target(0, 1, (0, 0, 0), (1, 1, 1)):
            if clearDest:
                gl.glClear(gl.GL_COLOR_BUFFER_BIT |
                           gl.GL_DEPTH_BUFFER_BIT |
                           gl.GL_STENCIL_BUFFER_BIT)
            self.apply(source, 0.5, 0, 1, 0, 1, 0, 1, **kwargs)



#!/usr/bin/env python
#
# plugins.py - Accessing installed FSLeyes plugins.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides access to installed FSLeyes plugins.

FSLeyes uses a simple plugin architecture for loading custom views, controls,
and tools. Plugins can be installed from Python libraries (e.g. as hosted on
`PyPi <https://pypi.org/>`_), or installed directly from a ``.py`` file.

In both cases, FSLeyes uses ``setuptools`` `entry points
<https://setuptools.readthedocs.io/en/latest/pkg_resources.html#entry-points>`__
to locate the items provided by plugin library/files.


Things plugins can provide
--------------------------


FSLeyes plugins can provide custom *views*, *controls* and *tools*:

 - A *view* is a top level panel, such as an :class:`.OrthoPanel`,
   :class:`.Scene3DPanel`, or :class:`.TimeSeriesPanel`. Views provided
   by plugins are added to the top level *Views* menu.

 - A *control* is a secondary panel, or toolbar, which is embedded within a
   view, such as an :class:`.OverlayListPanel`, :class:`.OrthoToolBar`, or
   :class:`.MelodicClassificationPanel`. Controls provided by plugins are
   added to the *Settings* menu for each active view.

 - A *tool* is an :class:`.Action` which is associated with a menu item
   under the top-level *Tools* menu, such as the :class:`.ApplyFlirtXfmAction`
   and the :class:`.ResampleAction`.


Loading/installing FSLeyes plugins
----------------------------------


FSLeyes plugins are loaded into a running FSLeyes as follows:

 - Any Python libraries (e.g. installed from ``PyPi``) which are present the
   environment that FSLeyes is running in, and which have a name beginning
   with ``fsleyes-plugin-`` will automatically be detected by FSLeyes.

 - Plugin ``.py`` files, which contain view, control, and/or tool definitions,
   can be passed directly to the :func:`loadPlugin` function.

 - Plugin ``.py`` files which are present in the FSLeyes settings directory,
   or which are found in the ``FSLEYES_PLUGIN_PATH`` environment variable, will
   be loaded by the :func:`initialise` function.


A plugin can be installed permanently into FSLeyes as follows:


 - Any Python libraries (e.g. installed from ``PyPi``) which are present the
   environment that FSLeyes is running in, and which have a name beginning
   with ``fsleyes-plugin-`` will automatically be detected by FSLeyes.

 - ``.py`` plugin files can be passed to the :func:`installPlugin`
   function. This file will be saved into the FSLeyes settings directory
   (e.g. ``~/.fsleyes/plugins/``).


Writing a FSLeyes plugin
------------------------


.. note:: A minimal example of a FSLeyes plugin library can be found in
          ``tests/testdata/fsleyes_plugin_example/``.


A FSLeyes plugin is a Python library, or a ``.py`` file, which contains
definitions for custom views, controls, and tools.

 - Views must be sub-classes of the :class:`.ViewPanel` class.

 - Controls must be sub-classes of the :class:`.ControlPanel` class. If your
   custom control is designed to only work with a specific view, you should
   override the :mod:`.ControlMixin.supportedViews` static method to return
   the views that your control supports.

 - Tools must be sub-classes of the :class:`.Action` class.


To write a ``.py`` file which can be loaded as a FSLeyes plugin, simply
define your views, controls, and tools in the file. The file path can then
be passed to the :func:`loadPlugin` or :func:`installPlugin` function.


To release a FSLeyes plugin as a library, you need to organise your code
as a Python library. Minimally, this requires the following:

 - Arrange your ``.py`` file(s) into a Python package.

 - Write a ``setup.py`` file.

 - Give your library a name (the ``name`` argument to the ``setup``
   function) which begins with ``fsleyes-plugin-``.

 - Expose your custom views, controls, and tools as `entry points
   <https://packaging.python.org/specifications/entry-points/>`__ (the
   ``entry_points`` argument to the ``setup`` function).


A minimal ``setup.py`` file for a FSLeyes plugin might look like this::

    import setuptools

    setup(
        # the name must begin with "fsleyes-plugin-"
        name='fsleyes-plugin-my-cool-plugin',

        # Views, controls, and tools must be exposed
        # as entry points within groups called
        # "fsleyes_views", "fsleyes_controls" and
        # "fsleyes_tools" respectively.
        entry_points={
            'fsleyes_views' : [
                'My cool view = myplugin:MyView'
            ]
            'fsleyes_controls' : [
                'My cool control = myplugin:MyControl'
            ]
            'fsleyes_tools' : [
                'My cool tool = myplugin:MyTool'
            ]
        }
    )


See the `Python Packaging guide
<https://packaging.python.org/tutorials/packaging-projects/>`_ for more
details on writing a ``setup.py`` file.


Module contents
---------------


The following functions can be used to load/install new plugins:

.. autosummary::
   :nosignatures:

   initialise
   loadPlugin
   installPlugin


The following functions can be used to access plugins:

.. autosummary::
   :nosignatures:

   listPlugins
   listViews
   listControls
   listTools

"""


import os.path as op
import            os
import            sys
import            glob
import            logging
import            collections
import            pkg_resources

import fsl.utils.settings            as fslsettings
import fsleyes.actions               as actions
import fsleyes.views.viewpanel       as viewpanel
import fsleyes.controls.controlpanel as ctrlpanel


log = logging.getLogger(__name__)


def initialise():
    """Calls :func:`loadPlugin` on all plugin files in the FSLeyes settings
    directory, and found on the ``FSLEYES_PLUGIN_PATH`` environment variable.
    """
    pluginFiles = list(fslsettings.listFiles('plugins/*.py'))
    pluginFiles = [fslsettings.filePath(p) for p in pluginFiles]
    fpp         = os.environ.get('FSLEYES_PLUGIN_PATH', None)

    if fpp is not None:
        for dirname in fpp.split(op.pathsep):
            pluginFiles.extend(glob.glob(op.join(dirname, '*.py')))

    for fname in pluginFiles:
        try:
            loadPlugin(fname)
        except Exception as e:
            log.warning('Failed to load plugin file %s: %s', fname, e)


def listPlugins():
    """Returns a list containing the names of all installed FSLeyes plugins.
    """
    plugins = []
    for dist in pkg_resources.working_set:
        if dist.project_name.startswith('fsleyes-plugin-'):
            plugins.append(dist.project_name)
    return list(sorted(plugins))


def _listEntryPoints(group):
    """Returns a dictionary containing ``{name : type}`` entry points for the
    given entry point group.

    https://setuptools.readthedocs.io/en/latest/pkg_resources.html#entry-points
    """
    items = collections.OrderedDict()
    for plugin in listPlugins():
        for name, ep in pkg_resources.get_entry_map(plugin, group).items():
            if name in items:
                log.debug('Overriding entry point %s [%s] with entry point '
                          'of the same name from %s', name, group, plugin)
            items[name] = ep.load()
    return items


def listViews():
    """Returns a dictionary of ``{name : ViewPanel}`` mappings containing
    the custom views provided by all installed FSLeyes plugins.
    """
    views = _listEntryPoints('fsleyes_views')
    for name, cls in list(views.items()):
        if not issubclass(cls, viewpanel.ViewPanel):
            log.debug('Ignoring fsleyes_views entry point '
                      '{} - not a ViewPanel'.format(name))
            views.pop(name)
            continue
    return views


def listControls(viewType=None):
    """Returns a dictionary of ``{name : ControlPanel}`` mappings containing
    the custom controls provided by all installed FSLeyes plugins.

    :arg viewType: Sub-class of :class:`.ViewPanel` - if provided, only
                   controls which are compatible with this view type are
                   returned (as determined by
                   :meth:`.ControlMixin.supportedViews.`).
    """
    ctrls = _listEntryPoints('fsleyes_controls')

    for name, cls in list(ctrls.items()):
        if not issubclass(cls, (ctrlpanel.ControlPanel,
                                ctrlpanel.ControlToolBar)):
            log.debug('Ignoring fsleyes_controls entry point {} - '
                      'not a ControlPanel/ToolBar'.format(name))
            ctrls.pop(name)
            continue

        supported = cls.supportedViews()
        if viewType  is not None and \
           supported is not None and \
           viewType  not in supported:
            ctrls.pop(name)
            continue
    return ctrls


def listTools():
    """Returns a dictionary of ``{name : Action}`` mappings containing
    the custom tools provided by all installed FSLeyes plugins.
    """
    tools = _listEntryPoints('fsleyes_tools')
    for name, cls in list(tools.items()):
        if not issubclass(cls, actions.Action):
            log.debug('Ignoring fsleyes_tools entry point '
                      '{} - not an Action'.format(name))
            tools.pop(name)
            continue
    return tools


def _findEntryPoints(filename, modname):
    """Used by :func:`loadPlugin`. Imports the given Python file (setting the
    module name to ``modname``), and finds the FSLeyes entry points (views,
    controls, or tools) that are defined within.
    """

    log.debug('Importing %s as %s', filename, modname)

    entryPoints = collections.defaultdict(dict)
    pyver       = sys.version_info[:2]

    # Ugh.
    if pyver >= (3, 5):
        import importlib.util as imputil
        spec = imputil.spec_from_file_location(modname, filename)
        mod  = imputil.module_from_spec(spec)
        spec.loader.exec_module(mod)

    elif pyver == (3, 4):
        from importlib.machinery import SourceFileLoader
        mod = SourceFileLoader(modname, filename).load_module()

    else:
        import imp
        mod = imp.load_source(modname, filename)

    sys.modules[modname] = mod

    for name in dir(mod):

        item  = getattr(mod, name)
        group = None

        if not isinstance(item, type):
            continue

        # avoid module imports
        if item is viewpanel.ViewPanel    or \
           item is ctrlpanel.ControlPanel or \
           item is actions.Action:
            continue

        if   issubclass(item, viewpanel.ViewPanel):    group = 'views'
        elif issubclass(item, ctrlpanel.ControlPanel): group = 'controls'
        elif issubclass(item, actions.Action):         group = 'tools'

        if group is not None:
            log.debug('Found %s entry point: %s', group, name)
            entryPoints['fsleyes_{}'.format(group)][name] = item

    return entryPoints


def loadPlugin(filename):
    """Loads the given Python file as a FSLeyes plugin. """

    name     = op.splitext(op.basename(filename))[0]
    modname  = 'fsleyes_plugin_{}'.format(name)
    distname = 'fsleyes-plugin-{}'.format(name)

    if distname in listPlugins():
        log.debug('Plugin %s is already in environment - skipping', distname)
        return

    log.debug('Loading plugin %s [dist name %s]', filename, distname)

    dist = pkg_resources.Distribution(
        project_name=distname,
        location=op.dirname(filename),
        version='0.0.0')

    entryPoints = _findEntryPoints(filename, modname)

    # Here I'm relying on the fact that
    # Distribution.get_entry_map returns
    # the actual dict that it uses to
    # store entry points.
    entryMap = dist.get_entry_map()

    for group, entries in entryPoints.items():
        entryMap[group] = {}

        for name, item in entries.items():
            ep = '{} = {}:{}'.format(name, modname, name)
            ep = pkg_resources.EntryPoint.parse(ep, dist=dist)
            entryMap[group][name] = ep

    pkg_resources.working_set.add(dist)


def installPlugin(filename):
    """Copies the given Python file into the FSLeyes settings directory,
    within a sub-directory called ``plugins``. After the file has been
    copied, the path to the copy is passed to :func:`loadPlugin`.
    """

    basename = op.splitext(op.basename(filename))[0]
    dest     = 'plugins/{}.py'.format(basename)

    log.debug('Installing plugin %s', filename)

    with open(filename, 'rt')        as inf, \
         fslsettings.writeFile(dest) as outf:
        outf.write(inf.read())

    dest = fslsettings.filePath(dest)

    try:
        loadPlugin(dest)
    except Exception:
        fslsettings.deleteFile(dest)
        raise



#!/usr/bin/env python
#
# plottoolbar.py - The PlotToolBar class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`PlotToolBar` class, a toolbar for use with
an :class:`.OverlayPlotPanel`.
"""


import fsleyes_props    as props

import fsleyes.controls.controlpanel as ctrlpanel
import fsleyes.icons                 as icons
import fsleyes.actions               as actions
import fsleyes.tooltips              as tooltips


class PlotToolBar(ctrlpanel.ControlToolBar):
    """The ``PlotToolBar`` is a toolbar for use with an
    :class:`.OverlayPlotPanel`. It creates toolbar controls which
    are common to all :class:`.OverlayPlotPanel` types.
    """

    def __init__(self, parent, overlayList, displayCtx, frame, plotPanel):
        """Create a ``PlotToolBar``.

        :arg parent:      The :mod:`wx` parent object.
        :arg overlayList: The :class:`.OverlayList` instance.
        :arg displayCtx:  The :class:`.DisplayContext` instance.
        :arg frame:       The :class:`.FSLeyesFrame` instance.
        :arg plotPanel:   The :class:`.PlotPanel` instance.
        """

        ctrlpanel.ControlToolBar.__init__(self,
                                          parent,
                                          overlayList,
                                          displayCtx,
                                          frame,
                                          height=24,
                                          kbFocus=True)

        self.__plotPanel = plotPanel

        import_     = actions.ActionButton(
            'importDataSeries',
            icon=icons.findImageFile('importDataSeries24'),
            tooltip=tooltips.actions[plotPanel, 'importDataSeries'])
        export      = actions.ActionButton(
            'exportDataSeries',
            icon=icons.findImageFile('exportDataSeries24'),
            tooltip=tooltips.actions[plotPanel, 'exportDataSeries'])
        add        = actions.ActionButton(
            'addDataSeries',
            icon=icons.findImageFile('add24'),
            tooltip=tooltips.actions[plotPanel, 'addDataSeries'])
        remove     = actions.ActionButton(
            'removeDataSeries',
            icon=icons.findImageFile('remove24'),
            tooltip=tooltips.actions[plotPanel, 'removeDataSeries'])
        screenshot = actions.ActionButton(
            'screenshot',
            icon=icons.findImageFile('camera24'),
            tooltip=tooltips.actions[plotPanel, 'screenshot'])

        screenshot = props.buildGUI(self, plotPanel, screenshot)
        import_    = props.buildGUI(self, plotPanel, import_)
        export     = props.buildGUI(self, plotPanel, export)
        add        = props.buildGUI(self, plotPanel, add)
        remove     = props.buildGUI(self, plotPanel, remove)

        self.__commonTools = [screenshot, import_, export, add, remove]
        self.__commonNav   = [screenshot, import_, export, add, remove]

        self.SetTools([screenshot, import_, export, add, remove])


    def destroy(self):
        """Must be called when this ``PlotToolBar`` is no longer needed.
        Clears some references and calls the base class implementation.
        """
        self.__commonTools = None
        self.__commonNav   = None
        self.__plotPanel   = None

        ctrlpanel.ControlToolBar.destroy(self)


    def getCommonTools(self):
        """Returns a list containing the toolbar widgets added by this
        ``PlotToolBar``.
        """
        return list(self.__commonTools)


    def getCommonNavOrder(self):
        """Returns a list containing the navigation order for tools added
        by this ``PlotToolBar``.
        """
        return list(self.__commonNav)


    def getPlotPanel(self):
        """Returns the :class:`.OverlayPlotPanel` bound to this ``PlotToolBar``.
        """
        return self.__plotPanel
